import math
import random

import numpy as np
from numpy.core.umath_tests import inner1d

from neupy.utils import format_data
from neupy.core.properties import Property
from .base import DiscreteMemory


__all__ = ('DiscreteHopfieldNetwork',)


def bin2sign(matrix):
    return np.where(matrix == 0, -1, 1)


def hopfield_energy(weight, X, output_data):
    return -0.5 * inner1d(X.dot(weight), output_data)


class DiscreteHopfieldNetwork(DiscreteMemory):
    """
    Discrete Hopfield Network. It can memorize binary samples
    and reconstruct them from corrupted samples.

    Notes
    -----
    - Works only with binary data. Input matrix should
      contain only zeros and ones.

    Parameters
    ----------
    {DiscreteMemory.Parameters}

    check_limit : bool
        Option enable a limit of patterns control for the
        network using logarithmically proportion rule.
        Defaults to ``True``.

        .. math::

            \\frac{{n_{{features}}}}{{2 \\cdot log_{{e}}(n_{{features}})}}

    Methods
    -------
    energy(X)
        Computes Discrete Hopfield Energy.

    train(X)
        Save input data pattern into the network's memory. Each call will
        make partial fit for the network.

    predict(X, n_times=None)
        Recover data from the memory using input pattern.
        For the prediction procedure you can control number
        of iterations. If you set up this value equal to ``None``
        then the value would be equal to the value that you
        set up for the property with the same name - ``n_times``.

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> def draw_bin_image(image_matrix):
    ...     for row in image_matrix.tolist():
    ...         print('| ' + ' '.join(' *'[val] for val in row))
    ...
    >>> zero = np.matrix([
    ...     0, 1, 1, 1, 0,
    ...     1, 0, 0, 0, 1,
    ...     1, 0, 0, 0, 1,
    ...     1, 0, 0, 0, 1,
    ...     1, 0, 0, 0, 1,
    ...     0, 1, 1, 1, 0
    ... ])
    >>>
    >>> one = np.matrix([
    ...     0, 1, 1, 0, 0,
    ...     0, 0, 1, 0, 0,
    ...     0, 0, 1, 0, 0,
    ...     0, 0, 1, 0, 0,
    ...     0, 0, 1, 0, 0,
    ...     0, 0, 1, 0, 0
    ... ])
    >>>
    >>> two = np.matrix([
    ...     1, 1, 1, 0, 0,
    ...     0, 0, 0, 1, 0,
    ...     0, 0, 0, 1, 0,
    ...     0, 1, 1, 0, 0,
    ...     1, 0, 0, 0, 0,
    ...     1, 1, 1, 1, 1,
    ... ])
    >>>
    >>> half_zero = np.matrix([
    ...     0, 1, 1, 1, 0,
    ...     1, 0, 0, 0, 1,
    ...     1, 0, 0, 0, 1,
    ...     0, 0, 0, 0, 0,
    ...     0, 0, 0, 0, 0,
    ...     0, 0, 0, 0, 0,
    ... ])
    >>>
    >>> draw_bin_image(zero.reshape((6, 5)))
    |   * * *
    | *       *
    | *       *
    | *       *
    | *       *
    |   * * *
    >>> draw_bin_image(half_zero.reshape((6, 5)))
    |   * * *
    | *       *
    | *       *
    |
    |
    |
    >>> data = np.concatenate([zero, one, two], axis=0)
    >>>
    >>> dhnet = algorithms.DiscreteHopfieldNetwork()
    >>> dhnet.train(data)
    >>>
    >>> result = dhnet.predict(half_zero)
    >>> draw_bin_image(result.reshape((6, 5)))
    |   * * *
    | *       *
    | *       *
    | *       *
    | *       *
    |   * * *

    See Also
    --------
    :ref:`password-recovery`: Password recovery with Discrete Hopfield Network.
    :ref:`discrete-hopfield-network`: Discrete Hopfield Network article.
    """
    check_limit = Property(expected_type=bool)

    def __init__(self, mode='sync', n_times=100, verbose=False,
                 check_limit=True):

        self.n_memorized_samples = 0
        self.check_limit = check_limit

        super(DiscreteHopfieldNetwork, self).__init__(
            mode=mode, n_times=n_times, verbose=verbose)

    def train(self, X_bin):
        self.discrete_validation(X_bin)

        X_sign = bin2sign(X_bin)
        X_sign = format_data(X_sign, is_feature1d=False, make_float=False)

        n_rows, n_features = X_sign.shape
        n_rows_after_update = self.n_memorized_samples + n_rows

        if self.check_limit:
            memory_limit = math.ceil(n_features / (2 * math.log(n_features)))

            if n_rows_after_update > memory_limit:
                raise ValueError("You can't memorize more than {0} "
                                 "samples".format(memory_limit))

        weight_shape = (n_features, n_features)

        if self.weight is None:
            self.weight = np.zeros(weight_shape, dtype=int)

        if self.weight.shape != weight_shape:
            n_features_expected = self.weight.shape[1]
            raise ValueError("Input data has invalid number of features. "
                             "Got {} features instead of {}."
                             "".format(n_features, n_features_expected))

        self.weight += X_sign.T.dot(X_sign)
        np.fill_diagonal(self.weight, np.zeros(len(self.weight)))
        self.n_memorized_samples = n_rows_after_update

    def predict(self, X_bin, n_times=None):
        self.discrete_validation(X_bin)

        X_sign = bin2sign(X_bin)
        X_sign = format_data(X_sign, is_feature1d=False, make_float=False)

        if self.mode == 'async':
            _, n_features = X_sign.shape

            if n_times is None:
                n_times = self.n_times

            for _ in range(n_times):
                position = random.randrange(n_features)
                raw_new_value = X_sign.dot(self.weight[:, position])
                X_sign[:, position] = np.sign(raw_new_value)
        else:
            X_sign = X_sign.dot(self.weight)

        return np.where(X_sign > 0, 1, 0).astype(int)

    def energy(self, X_bin):
        self.discrete_validation(X_bin)

        X_sign = bin2sign(X_bin)
        X_sign = format_data(X_sign, is_feature1d=False, make_float=False)

        n_rows, n_features = X_sign.shape

        if n_rows == 1:
            return hopfield_energy(self.weight, X_sign, X_sign)

        output = np.zeros(n_rows)
        for i, row in enumerate(X_sign):
            output[i] = hopfield_energy(self.weight, row, row)

        return output



import tensorflow as tf

from neupy.core.properties import ProperFractionProperty, Property
from .base import GradientDescent


__all__ = ('Momentum',)


class Momentum(GradientDescent):
    """
    Momentum algorithm.

    Parameters
    ----------
    momentum : float
        Control previous gradient ratio. Defaults to ``0.9``.

    nesterov : bool
        Instead of classic momentum computes Nesterov momentum.
        Defaults to ``False``.

    {GradientDescent.Parameters}

    Attributes
    ----------
    {GradientDescent.Attributes}

    Methods
    -------
    {GradientDescent.Methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>> from neupy.layers import *
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> network = Input(2) >> Sigmoid(3) >> Sigmoid(1)
    >>> optimizer = algorithms.Momentum(network)
    >>> optimizer.train(x_train, y_train)

    See Also
    --------
    :network:`GradientDescent` : GradientDescent algorithm.
    """
    momentum = ProperFractionProperty(default=0.9)
    nesterov = Property(default=False, expected_type=bool)

    def init_train_updates(self):
        optimizer = tf.train.MomentumOptimizer(
            use_nesterov=self.nesterov,
            momentum=self.momentum,
            learning_rate=self.step,
        )
        self.functions.optimizer = optimizer
        return [optimizer.minimize(self.variables.loss)]



import copy
from functools import reduce

import tensorflow as tf

from neupy.core.properties import FunctionWithOptionsProperty, IntProperty
from neupy.exceptions import LayerConnectionError
from neupy.utils import tf_utils
from .base import BaseLayer


__all__ = ('Elementwise', 'Concatenate', 'GatedAverage')


class Elementwise(BaseLayer):
    """
    Layer merges multiple input with elementwise function and generate
    single output. Each input to this layer should have exactly the same
    shape, otherwise it won't be possible to apply elementwise operation.

    Parameters
    ----------
    merge_function : callable or {{``add``, ``multiply``}}
        Callable object that accepts two inputs and
        combines them in value using elementwise operation.

        - ``add`` - Sum all the inputs. Alias to ``tf.add``.

        - ``multiply`` - Multiplies all the inputs. Alias to ``tf.multiply``.

        - Custom function requires to have two input arguments.

        .. code-block:: python

            def subtraction(x, y):
                return x - y

        Defaults to ``add``.

    {BaseLayer.name}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}

    Examples
    --------
    >>> from neupy import layers
    >>> network = (Input(10) | Input(10)) >> Elementwise('add')
    [(?, 10), (?, 10)] -> [... 3 layers ...] -> (?, 10)
    """
    merge_function = FunctionWithOptionsProperty(choices={
        'add': tf.add,
        'multiply': tf.multiply,
    })

    def __init__(self, merge_function='add', name=None):
        super(Elementwise, self).__init__(name=name)
        self.original_function = merge_function
        self.merge_function = merge_function

    def get_output_shape(self, *input_shapes):
        input_shapes = [tf.TensorShape(shape) for shape in input_shapes]
        first_shape = input_shapes[0]

        if len(input_shapes) < 2:
            raise LayerConnectionError(
                "Layer `{}` expected multiple inputs. Input shapes: {}"
                "".format(self.name, tf_utils.shape_to_tuple(input_shapes)))

        if any(shape.ndims is None for shape in input_shapes):
            return tf.TensorShape(None)

        for shape in input_shapes:
            if not shape.is_compatible_with(first_shape):
                formatted_shapes = tf_utils.shape_to_tuple(input_shapes)
                raise LayerConnectionError(
                    "Input shapes to the `{}` layer have incompatible shapes. "
                    "Input shapes: {}, Layer: {}"
                    "".format(self.name, formatted_shapes, self))

        return first_shape

    def output(self, *inputs, **kwargs):
        return reduce(self.merge_function, inputs)

    def __repr__(self):
        return self._repr_arguments(
            repr(self.original_function), name=self.name)


class Concatenate(BaseLayer):
    """
    Concatenate multiple inputs into one. Inputs will be concatenated over
    the specified axis (controlled with parameter ``axis``).

    Parameters
    ----------
    axis : int
        The axis along which the inputs will be concatenated.
        Default is ``-1``.

    {BaseLayer.name}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}

    Examples
    --------
    >>> from neupy.layers import *
    >>> network = (Input(10) | Input(20)) >> Concatenate()
    [(?, 10), (?, 20)] -> [... 3 layers ...] -> (?, 30)
    """
    axis = IntProperty()

    def __init__(self, axis=-1, name=None):
        super(Concatenate, self).__init__(name=name)
        self.axis = axis

    def get_output_shape(self, *input_shapes):
        input_shapes = [tf.TensorShape(shape) for shape in input_shapes]
        # The axis value has 0-based indices where 0s index points
        # to the batch dimension of the input. Shapes in the neupy
        # do not store information about the batch and we need to
        # put None value on the 0s position.
        valid_shape = input_shapes[0]

        if any(shape.ndims is None for shape in input_shapes):
            return tf.TensorShape(None)

        # Avoid using negative indices
        possible_axes = list(range(len(valid_shape)))
        concat_axis = possible_axes[self.axis]

        for input_shape in input_shapes[1:]:
            if len(valid_shape) != len(input_shape):
                raise LayerConnectionError(
                    "Cannot concatenate layers, because inputs have "
                    "different number of dimensions. Shapes: {} and {}"
                    "".format(valid_shape, input_shape))

            for axis, axis_size in enumerate(input_shape):
                if axis != concat_axis and valid_shape[axis] != axis_size:
                    raise LayerConnectionError(
                        "Cannot concatenate layers, because some of them "
                        "don't match over dimension #{} (0-based indices). "
                        "Shapes: {} and {}"
                        "".format(axis, valid_shape, input_shape))

        output_shape = input_shapes.pop(0)
        output_shape = [dim.value for dim in output_shape.dims]

        for input_shape in input_shapes:
            output_shape[self.axis] += input_shape[self.axis]

        return tf.TensorShape(output_shape)

    def output(self, *inputs, **kwargs):
        return tf.concat(inputs, axis=self.axis)


def exclude_index(array, index):
    """
    Copies array and exclude single element in specific
    index position.

    Parameters
    ----------
    array : list or tuple

    index : int
        Index of the value that has to be excluded from the array

    Returns
    -------
    list
    """
    array = list(array)
    copied_array = copy.copy(array)
    copied_array.pop(index)
    return copied_array


class GatedAverage(BaseLayer):
    """
    Layer uses applies weighted elementwise addition to multiple outputs.
    Weight can be control using separate input known as **gate**. Number
    of outputs from the gate has to be equal to the number of networks,
    since each value from the weight will be a weight per each network.

    Layer expects gate as a first input, but it can be controlled with
    the ``gate_index`` parameter.

    Parameters
    ----------
    gate_index : int
        Input layers passed as a list and current variable specifies
        index in which it can find gating network. Defaults to ``0``,
        which means that it expects to see gating layer in first position.

    {BaseLayer.Parameters}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}

    Examples
    --------
    >>> from neupy.layers import *
    >>>
    >>> gate = Input(10) >> Softmax(2)
    >>> net1 = Input(20) >> Relu(10)
    >>> net2 = Input(20) >> Relu(20) >> Relu(10)
    >>>
    >>> network = (gate | net1 | net2) >> GatedAverage()
    >>> network
    [(10,), (20,), (20,)] -> [... 8 layers ...] -> 10
    """
    gate_index = IntProperty(default=0)

    def __init__(self, gate_index=0, name=None):
        super(GatedAverage, self).__init__(name=name)
        self.gate_index = gate_index

    def fail_if_shape_invalid(self, input_shapes):
        n_input_layers = len(input_shapes)

        try:
            gate_shape = input_shapes[self.gate_index]
        except IndexError:
            raise LayerConnectionError(
                "Invalid index for gating layer. Number of input "
                "layers: {}. Gating layer index: {}"
                "".format(n_input_layers, self.gate_index))

        other_shapes = exclude_index(input_shapes, self.gate_index)
        if gate_shape and len(gate_shape) != 2:
            raise LayerConnectionError(
                "Output from the gating network should be 2-dimensional. "
                "Output shape from gating layer: {!r}"
                "".format(gate_shape))

        n_expected_networks = gate_shape[-1]
        # Note: -1 from all layers in order to exclude gating layer
        if n_expected_networks != (n_input_layers - 1):
            raise LayerConnectionError(
                "Gating layer can work only for combining only {} networks, "
                "got {} networks instead."
                "".format(n_expected_networks, (n_input_layers - 1)))

        for shape in other_shapes:
            if not shape.is_compatible_with(other_shapes[0]):
                raise LayerConnectionError(
                    "Output layer that has to be merged expect to "
                    "have the same shapes. Shapes: {!r}"
                    "".format(tf_utils.shape_to_tuple(other_shapes)))

    def get_output_shape(self, *input_shapes):
        input_shapes = [tf.TensorShape(shape) for shape in input_shapes]

        if any(shape.ndims is None for shape in input_shapes):
            return tf.TensorShape(None)

        self.fail_if_shape_invalid(input_shapes)

        if self.gate_index >= 0:
            # Take layer from the left side from the gating layer.
            # In case if gating layer at the zeros position then
            # it will take the last layer (-1 index).
            return input_shapes[self.gate_index - 1]

        # In case if it negative index, we take layer from the right side
        return input_shapes[self.gate_index + 1]

    def output(self, *inputs, **kwargs):
        gating_value = inputs[self.gate_index]
        other_values = exclude_index(inputs, self.gate_index)
        output_values = []

        for i, other_value in enumerate(other_values):
            n_feature_dim = other_value.shape.ndims - 1
            gate = tf.reshape(gating_value[:, i], [-1] + [1] * n_feature_dim)
            output_value = tf.multiply(other_value, gate)
            output_values.append(output_value)

        return sum(output_values)



from __future__ import division

import numpy as np

from neupy.core.docs import shared_docs


def gaussian_df(data, mean=0, std=1):
    """
    Returns gaussian density for each data sample.
    Gaussian specified by the mean and standard deviation.

    Parameters
    ----------
    data : array-like

    mean : float
        Gaussian mean.

    std : float
        Gaussian standard deviation.
    """
    if std == 0:
        return np.where(data == mean, 1, 0)

    normalizer = 2 * np.pi * std ** 2
    return np.exp(-np.square(data - mean) / normalizer)


def find_neighbour_distance(grid, center):
    """
    Returns distance from the center into different directions
    per each dimension separately.

    Parameters
    ----------
    grid : array-like
       Array that contains grid of n-dimensional vectors.

    center : tuple
        Index of the main neuron for which function returns
        distance to neuron's neighbours.

    Returns
    -------
    list of n-dimensional vectors
    """
    if len(center) != grid.ndim:
        raise ValueError(
            "Cannot find center, because grid of neurons has {} dimensions "
            "and center has specified coordinates for {} dimensional grid"
            "".format(grid.ndim, len(center)))

    slices = []
    for dim_length, center_coord in zip(grid.shape, center):
        slices.append(slice(-center_coord, dim_length - center_coord))

    return np.ogrid[slices]


@shared_docs(find_neighbour_distance)
def find_step_scaler_on_rect_grid(grid, center, std=1):
    """
    Function returns multivariate gaussian around the center
    with specified standard deviation.

    Parameters
    ----------
    {find_neighbour_distance.grid}

    {find_neighbour_distance.center}

    std : int, float
        Gaussian standard deviation. Defaults to ``1``.
    """
    distances = find_neighbour_distance(grid, center)
    gaussian_array = sum(gaussian_df(dist, std=std) for dist in distances)
    return gaussian_array / grid.ndim


@shared_docs(find_neighbour_distance)
def find_neighbours_on_rect_grid(grid, center, radius):
    """
    Function find all neuron's neighbours around specified
    center within a certain radius.

    Parameters
    ----------
    {find_neighbour_distance.grid}

    {find_neighbour_distance.center}

    radius : int
        Radius specifies what neurons, around the center, are neighbours.

    Returns
    -------
    array-like
        Return matrix with the same dimension as ``grid``
        where center element and it neighbours positions
        filled with value ``1`` and other as a ``0`` value.

    Examples
    --------
    >>> import numpy as np
    >>> from neupy.algorithms.competitive import sofm
    >>>
    >>> sofm.find_neighbours_on_rect_grid(
    ...     grid=np.zeros((3, 3)),
    ...     center=(0, 0),
    ...     radius=1)
    ...
    array([[ 1.,  1.,  0.],
           [ 1.,  0.,  0.],
           [ 0.,  0.,  0.]])
    >>>
    >>> sofm.find_neighbours_on_rect_grid(
    ...     grid=np.zeros((5, 5)),
    ...     center=(2, 2),
    ...     radius=2)
    ...
    array([[ 0.,  0.,  1.,  0.,  0.],
           [ 0.,  1.,  1.,  1.,  0.],
           [ 1.,  1.,  1.,  1.,  1.],
           [ 0.,  1.,  1.,  1.,  0.],
           [ 0.,  0.,  1.,  0.,  0.]])
    """
    distances = find_neighbour_distance(grid, center)
    mask = sum(dist ** 2 for dist in distances) <= radius ** 2
    grid[mask] = 1
    return grid


def generate_neighbours_pattern(radius):
    """
    Generates pattern that defines neighbours
    within specified radius on the hexagon grid.

    Parameters
    ----------
    radius : int

    Returns
    -------
    array-like
        2d-array that has number of rows equal to number
        of columns equal to ``2 * radius + 1``
    """
    cache = generate_neighbours_pattern.cache

    if radius in cache:
        return cache[radius]

    size = 2 * radius + 1
    max_distance = radius + 1
    pattern = np.zeros((size, size))

    for i, width in enumerate(range(size - radius, size + 1)):
        start_point = radius - int(np.floor(width / 2))
        neighbour_range = slice(start_point, start_point + width)

        neighbour_distance = np.pad(
            array=(max_distance - i) * [max_distance - i],
            pad_width=i,
            mode='linear_ramp',
            end_values=max_distance)

        pattern[i, neighbour_range] = neighbour_distance
        # every row in the pattern is mirror symmetric
        pattern[size - i - 1, neighbour_range] = neighbour_distance

    cache[radius] = pattern
    return pattern


generate_neighbours_pattern.cache = {}


def find_neighbours_on_hexagon_grid(grid, center, radius):
    """
    Marks neighbors on the hexagon grid around specified
    center within specified radius.

    Parameters
    ----------
    grid : 2d-arry
        Hexagon grid.

    center : tuple
        Coordinates of the center neuron on the grid.
        Should be a tuple with two integers: ``(x, y)``.

    radius : int
        Radius specifies what neurons, around the center, are neighbours.

    Returns
    -------
    2d-array
    """
    if radius == 0:
        grid[center] = 1
        return grid

    center_x, center_y = center
    pattern = generate_neighbours_pattern(radius)

    # We add zero padding in order to be able
    # insert full pattern into the matrix. It's
    # useful in case if we are not able to mark all
    # neighbours from the pattern for specific neuron.
    # For instance, in case when center = (0, 0) we
    # are not able to mark neurons on the left and top
    # sides of the pattern's center.
    grid = np.pad(grid, radius, mode='constant')

    if center_x % 2 == 1:
        # Since every even row shifted to the right
        # compare to the odd rows then odd rows shifted
        # to the left compare to the even rows.
        # It means that our pattern has to be reversed
        # for the even rows.
        pattern = pattern[:, ::-1]

    pattern_size = 2 * radius + 1
    x_range = slice(center_x, center_x + pattern_size)
    y_range = slice(center_y, center_y + pattern_size)

    grid[x_range, y_range] = pattern

    # removing padding
    return grid[radius:-radius, radius:-radius]


@shared_docs(find_step_scaler_on_rect_grid)
def find_step_scaler_on_hexagon_grid(grid, center=None, std=1):
    """
    Find step scale for neighbouring neurons. The further
    neuron from the center the lower step scale it has.

    Notes
    -----
    Non-neighbor neurons will have high step size
    We do not set up it to zero, since later we will
    ignore these values anyway

    Parameters
    ----------
    grid : array-like

    center : object
        Variable is going to be ignored. It's defined
        only for compatibility with other functions

    {find_step_scaler_on_rect_grid.std}
    """
    return gaussian_df(grid, mean=1, std=std)



import tensorflow as tf

from neupy.core.properties import (
    ProperFractionProperty,
    NumberProperty, Property,
)
from .base import GradientDescent


__all__ = ('RMSProp',)


class RMSProp(GradientDescent):
    """
    RMSProp algorithm.

    Parameters
    ----------
    decay : float
        Decay rate. Value need to be between ``0`` and ``1``.
        Defaults to ``0.95``.

    momentum : float
        Defaults to ``0``.

    epsilon : float
        Value need to be greater than ``0``. Defaults to ``1e-7``.

    centered : bool
        (from Tensorflow documentation) If ``True``, gradients are
        normalized by the estimated variance of the gradient; if ``False``,
        by the uncentered second moment. Setting this to ``True`` may
        help with training, but is slightly more expensive in terms
        of computation and memory. Defaults to ``False``.

    {GradientDescent.Parameters}

    Attributes
    ----------
    {GradientDescent.Attributes}

    Methods
    -------
    {GradientDescent.Methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>> from neupy.layers import *
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> network = Input(2) >> Sigmoid(3) >> Sigmoid(1)
    >>> optimizer = algorithms.RMSProp(network)
    >>> optimizer.train(x_train, y_train)
    """
    decay = ProperFractionProperty(default=0.95)
    momentum = NumberProperty(default=0, minval=0)
    epsilon = NumberProperty(default=1e-7, minval=0)
    centered = Property(default=False, expected_type=bool)

    def init_train_updates(self):
        optimizer = tf.train.RMSPropOptimizer(
            decay=self.decay,
            momentum=self.momentum,
            centered=self.centered,
            epsilon=self.epsilon,
            learning_rate=self.step,
        )
        self.functions.optimizer = optimizer
        return [optimizer.minimize(self.variables.loss)]



import numpy as np
from sklearn import datasets
import matplotlib.pyplot as plt

from neupy.layers import *
from neupy import algorithms


def load_data():
    X, _ = datasets.fetch_openml('mnist_784', version=1, return_X_y=True)
    X = (X / 255.).astype(np.float32)

    np.random.shuffle(X)
    x_train_2d, x_test_2d = X[:60000], X[60000:]
    x_train_4d = x_train_2d.reshape((60000, 28, 28, 1))
    x_test_4d = x_test_2d.reshape((10000, 28, 28, 1))

    return x_train_4d, x_test_4d


def visualize_reconstructions(x_test_4d, n_samples=6):
    x_test = x_test_4d.reshape(x_test_4d.shape[0], -1)
    images = x_test[:n_samples] * 255.
    predicted_images = optimizer.predict(x_test_4d[:n_samples])
    predicted_images = predicted_images * 255.

    # Compare real and reconstructed images
    fig, axes = plt.subplots(n_samples, 2, figsize=(12, 8))
    iterator = zip(axes, images, predicted_images)

    for (left_ax, right_ax), real_image, predicted_image in iterator:
        real_image = real_image.reshape((28, 28))
        predicted_image = predicted_image.reshape((28, 28))

        left_ax.imshow(real_image, cmap=plt.cm.binary)
        right_ax.imshow(predicted_image, cmap=plt.cm.binary)

    plt.show()


if __name__ == '__main__':
    network = join(
        Input((28, 28, 1)),

        Convolution((3, 3, 16)) >> Relu(),
        Convolution((3, 3, 16)) >> Relu(),
        MaxPooling((2, 2)),

        Convolution((3, 3, 32)) >> Relu(),
        MaxPooling((2, 2)),

        Reshape(),

        Relu(128),
        Relu(16),

        # Notice that in the decoder every operation reverts back
        # changes from the encoder layer.
        Relu(128),

        # 800 is a shape that we got after we reshaped our image in the
        # Reshape layer
        Relu(800),

        Reshape((5, 5, 32)),

        # Upscaling layer reverts changes from the max pooling layer
        Upscale((2, 2)),

        # Deconvolution (a.k.a Transposed Convolution) reverts
        # changes done by Convolution
        Deconvolution((3, 3, 16)) >> Relu(),

        Upscale((2, 2)),
        Deconvolution((3, 3, 16)) >> Relu(),
        Deconvolution((3, 3, 1)) >> Sigmoid()
    )
    optimizer = algorithms.Momentum(
        network,
        step=0.02,
        momentum=0.9,
        batch_size=128,
        loss='rmse',

        shuffle_data=True,
        verbose=True,

        regularizer=algorithms.l2(0.01),
    )

    x_train_4d, x_test_4d = load_data()
    optimizer.train(x_train_4d, x_train_4d, x_test_4d, x_test_4d, epochs=1)
    visualize_reconstructions(x_test_4d, n_samples=6)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import time

import progressbar

from neupy.utils import iters


__all__ = ('ProgressbarSignal', 'PrintLastErrorSignal', 'EpochEndSignal',
           'ErrorCollector')


def format_time(time):
    """
    Format seconds into human readable format.

    Parameters
    ----------
    time : float
        Time specified in seconds

    Returns
    -------
    str
        Formated time.
    """
    mins, seconds = divmod(int(time), 60)
    hours, minutes = divmod(mins, 60)

    if hours > 0:
        return '{:0>2d}:{:0>2d}:{:0>2d}'.format(hours, minutes, seconds)

    elif minutes > 0:
        return '{:0>2d}:{:0>2d}'.format(minutes, seconds)

    elif seconds > 0:
        return '{:.0f} sec'.format(seconds)

    elif time >= 1e-3:
        return '{:.0f} ms'.format(time * 1e3)

    elif time >= 1e-6:
        return '{:.0f} μs'.format(time * 1e6)  # microseconds

    return '{:.0f} ns'.format(time * 1e9)  # nanoseconds or smaller


class EpochEndSignal(object):
    def __init__(self, function):
        self.function = function

    def epoch_end(self, network):
        self.function(network)


class ProgressbarSignal(object):
    def train_start(self, network, **kwargs):
        if kwargs['batch_size'] is None:
            self.n_batches = 1
            return

        self.n_batches = iters.count_minibatches(
            kwargs['X_train'],
            kwargs['batch_size'])

    def epoch_start(self, network):
        self.index = 0
        self.last_train_error = None
        self.bar = progressbar.NullBar()

        if network.verbose and self.n_batches >= 2:
            self.bar = iters.make_progressbar(self.n_batches, show_output=True)
            self.bar.update(0)

    def train_error(self, network, **data):
        self.last_train_error = data['value']

    def update_end(self, network):
        self.index += 1
        self.bar.update(self.index, loss=self.last_train_error)

    def epoch_end(self, network):
        self.bar.finish(end='\r')
        self.bar.fd.write('\r' + ' ' * self.bar.term_width + '\r')

    def __reduce__(self):
        return self.__class__, tuple()


class PrintLastErrorSignal(object):
    def print_last_error(self, network):
        messages = []
        base_message = "#{} : [{}] ".format(
            network.last_epoch,
            format_time(self.last_epoch_time))

        if self.last_train_error is not None:
            messages.append("train: {:.6f}".format(self.last_train_error))

        if self.last_valid_error is not None:
            messages.append("valid: {:.6f}".format(self.last_valid_error))

        network.logs.write(base_message + ', '.join(messages))

    def train_start(self, network, **kwargs):
        self.first_epoch = network.last_epoch + 1
        self.last_epoch_shown = 0

    def epoch_start(self, network):
        self.epoch_start_time = time.time()
        self.last_train_error = None
        self.last_valid_error = None

    def train_error(self, network, value, **kwargs):
        self.last_train_error = value

    def valid_error(self, network, value, **kwargs):
        self.last_valid_error = value

    def epoch_end(self, network):
        epoch = network.last_epoch
        self.last_epoch_time = time.time() - self.epoch_start_time

        if epoch % network.show_epoch == 0 or epoch == self.first_epoch:
            self.print_last_error(network)
            self.last_epoch_shown = network.last_epoch

    def train_end(self, network):
        if network.last_epoch != self.last_epoch_shown:
            self.print_last_error(network)


class ErrorCollector(object):
    def __init__(self):
        self.valid = []
        self.train = []

    def train_error(self, network, value, **kwargs):
        self.train.append(value)

    def valid_error(self, network, value, **kwargs):
        self.valid.append(value)



from .base import *
from .graph import *
from .activations import *
from .convolutions import *
from .pooling import *
from .stochastic import *
from .normalization import *
from .merge import *
from .reshape import *
from .embedding import *
from .recurrent import *


# Extra aliases for the layers
MaxPool = MaxPooling
AvgPool = AveragePooling
GlobalPool = GlobalPooling

Conv = Convolution
BN = BatchNorm
GN = GroupNorm



import tensorflow as tf

from neupy.core.properties import (
    ProperFractionProperty,
    ScalarVariableProperty,
    NumberProperty,
)
from .base import GradientDescent


__all__ = ('Adadelta',)


class Adadelta(GradientDescent):
    """
    Adadelta algorithm.

    Parameters
    ----------
    rho : float
        Decay rate. Value need to be between ``0``
        and ``1``. Defaults to ``0.95``.

    epsilon : float
        Value need to be greater than ``0``. Defaults to ``1e-7``.

    step : float
        Learning rate, defaults to ``1.0``. Original paper doesn't have
        learning rate specified in the paper. Step value equal to ``1.0``
        allow to achieve the same effect, since multiplication by one won't
        have any effect on the update.

    {GradientDescent.batch_size}

    {BaseOptimizer.regularizer}

    {BaseOptimizer.network}

    {BaseOptimizer.loss}

    {BaseNetwork.show_epoch}

    {BaseNetwork.shuffle_data}

    {BaseNetwork.signals}

    Attributes
    ----------
    {GradientDescent.Attributes}

    Methods
    -------
    {GradientDescent.Methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>> from neupy.layers import *
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> network = Input(2) >> Sigmoid(3) >> Sigmoid(1)
    >>> optimizer = algorithms.Adadelta(network)
    >>> optimizer.train(x_train, y_train)

    References
    ----------
    [1] Matthew D. Zeiler,
        ADADELTA: An Adaptive Learning Rate Method
        https://arxiv.org/pdf/1212.5701.pdf
    """
    step = ScalarVariableProperty(default=1.0)
    rho = ProperFractionProperty(default=0.95)
    epsilon = NumberProperty(default=1e-7, minval=0)

    def init_train_updates(self):
        optimizer = tf.train.AdadeltaOptimizer(
            rho=self.rho,
            epsilon=self.epsilon,
            learning_rate=self.step,
        )
        self.functions.optimizer = optimizer
        return [optimizer.minimize(self.variables.loss)]



import argparse

import numpy as np
import matplotlib.pyplot as plt
from sklearn import datasets, preprocessing
from neupy import algorithms, utils

from utils import iter_neighbours


plt.style.use('ggplot')
utils.reproducible()

parser = argparse.ArgumentParser()
parser.add_argument('--expanded-heatmap', action='store_true')

class_parameters = [
    dict(
        marker='o',
        markeredgecolor='#E24A33',
        markersize=11,
        markeredgewidth=2,
        markerfacecolor='None',
    ),
    dict(
        marker='s',
        markeredgecolor='#348ABD',
        markersize=14,
        markeredgewidth=2,
        markerfacecolor='None',
    ),
]


def load_data():
    data, target = datasets.load_breast_cancer(return_X_y=True)

    scaler = preprocessing.MinMaxScaler()
    data = scaler.fit_transform(data)

    return data, target


def compute_heatmap(weight):
    heatmap = np.zeros((GRID_HEIGHT, GRID_WIDTH))
    for (neuron_x, neuron_y), neighbours in iter_neighbours(weight):
        total_distance = 0

        for (neigbour_x, neigbour_y) in neighbours:
            neuron_vec = weight[:, neuron_x, neuron_y]
            neigbour_vec = weight[:, neigbour_x, neigbour_y]

            distance = np.linalg.norm(neuron_vec - neigbour_vec)
            total_distance += distance

        avg_distance = total_distance / len(neighbours)
        heatmap[neuron_x, neuron_y] = avg_distance

    return heatmap


def compute_heatmap_expanded(weight):
    heatmap = np.zeros((2 * GRID_HEIGHT - 1, 2 * GRID_WIDTH - 1))
    for (neuron_x, neuron_y), neighbours in iter_neighbours(weight):
        for (neigbour_x, neigbour_y) in neighbours:
            neuron_vec = weight[:, neuron_x, neuron_y]
            neigbour_vec = weight[:, neigbour_x, neigbour_y]

            distance = np.linalg.norm(neuron_vec - neigbour_vec)

            if neuron_x == neigbour_x and (neigbour_y - neuron_y) == 1:
                heatmap[2 * neuron_x, 2 * neuron_y + 1] = distance

            elif (neigbour_x - neuron_x) == 1 and neigbour_y == neuron_y:
                heatmap[2 * neuron_x + 1, 2 * neuron_y] = distance

    return heatmap


if __name__ == '__main__':
    args = parser.parse_args()

    GRID_HEIGHT = 20
    GRID_WIDTH = 20

    sofm = algorithms.SOFM(
        n_inputs=30,
        features_grid=(GRID_HEIGHT, GRID_WIDTH),

        learning_radius=4,
        reduce_radius_after=50,

        step=0.5,
        std=1,

        shuffle_data=True,
        verbose=True,
    )

    data, target = load_data()
    sofm.train(data, epochs=300)
    clusters = sofm.predict(data).argmax(axis=1)

    plt.figure(figsize=(13, 13))
    plt.title("Embedded 30-dimensional dataset using SOFM")

    for actual_class, cluster_index in zip(target, clusters):
        cluster_x, cluster_y = divmod(cluster_index, GRID_HEIGHT)
        parameters = class_parameters[actual_class]

        if args.expanded_heatmap:
            plt.plot(2 * cluster_x, 2 * cluster_y, **parameters)
        else:
            plt.plot(cluster_x, cluster_y, **parameters)

    weight = sofm.weight.reshape((sofm.n_inputs, GRID_HEIGHT, GRID_WIDTH))

    if args.expanded_heatmap:
        heatmap = compute_heatmap_expanded(weight)
    else:
        heatmap = compute_heatmap(weight)

    plt.imshow(heatmap, cmap='Greys_r', interpolation='nearest')

    plt.axis('off')
    plt.colorbar()
    plt.show()



import sys
import copy
import tempfile
from itertools import chain
from functools import wraps
from abc import abstractmethod
from collections import OrderedDict

import six
import graphviz
import numpy as np
import tensorflow as tf

from neupy.core.config import ConfigurableABC, DumpableObject
from neupy.exceptions import LayerConnectionError
from neupy.utils import as_tuple, tf_utils, iters


__all__ = (
    'BaseGraph', 'LayerGraph',
    'join', 'parallel', 'merge', 'repeat',
)


def make_one_if_possible(shape):
    """
    Format layer's input or output shape.

    Parameters
    ----------
    shape : int or tuple

    Returns
    -------
    int or tuple
    """
    if isinstance(shape, (tuple, list)) and len(shape) == 1:
        return shape[0]
    return shape


def filter_graph(dictionary, include_keys):
    """
    Create new list that contains only values
    specified in the ``include_keys`` attribute.

    Parameters
    ----------
    dictionary : dict
        Original dictionary

    include_keys : list or tuple
        Keys that will copied from original dictionary
        into a new one.

    Returns
    -------
    dict
    """
    filtered_dict = OrderedDict()

    for key, value in dictionary.items():
        if key in include_keys:
            filtered_dict[key] = [v for v in value if v in include_keys]

    return filtered_dict


def is_cyclic(graph):
    """
    Check if graph has cycles.

    Parameters
    ----------
    graph : dict

    Returns
    -------
    bool
        Return ``True`` if the directed graph has a cycle.

    Examples
    --------
    >>> is_cyclic({1: [2], 2: [3], 3: [1]})
    True
    >>> is_cyclic({1: [2], 2: [3], 3: [4]})
    False
    """
    path = set()
    visited = set()

    def visit(vertex):
        if vertex in visited:
            return False

        visited.add(vertex)
        path.add(vertex)

        for neighbour in graph.get(vertex, ()):
            if neighbour in path or visit(neighbour):
                return True

        path.remove(vertex)
        return False

    return any(visit(vertex) for vertex in graph)


def find_outputs_in_graph(graph):
    outputs = []

    for from_node, to_nodes in graph.items():
        if not to_nodes:
            outputs.append(from_node)

    return outputs


def topological_sort(graph):
    """
    Repeatedly go through all of the nodes in the graph, moving each of
    the nodes that has all its edges resolved, onto a sequence that
    forms our sorted graph. A node has all of its edges resolved and
    can be moved once all the nodes its edges point to, have been moved
    from the unsorted graph onto the sorted one.

    Parameters
    ----------
    graph : dict
        Dictionary that has graph structure.

    Raises
    ------
    RuntimeError
        If graph has cycles.

    Returns
    -------
    list
        List of nodes sorted in topological order.
    """
    if not graph:
        return []

    if is_cyclic(graph):
        raise RuntimeError(
            "Cannot apply topological sort to the graphs with cycles")

    sorted_nodes = []
    graph_unsorted = graph.copy()

    while graph_unsorted:
        for node, edges in list(graph_unsorted.items()):
            if all(edge not in graph_unsorted for edge in edges):
                del graph_unsorted[node]
                sorted_nodes.append(node)

    return sorted_nodes


def lazy_property(function):
    attr = '_lazy__' + function.__name__

    @property
    @wraps(function)
    def wrapper(self):
        if not hasattr(self, attr):
            setattr(self, attr, function(self))
        return getattr(self, attr)

    return wrapper


class BaseGraph(ConfigurableABC, DumpableObject):
    events = []

    def __init__(self, forward_graph=None):
        self.forward_graph = OrderedDict(forward_graph or [])

    @lazy_property
    def backward_graph(self):
        # First we copy all the nodes in order to
        # make sure that order stays the same
        backward = OrderedDict([(node, []) for node in self.forward_graph])

        for to_node, from_nodes in self.forward_graph.items():
            for from_node in from_nodes:
                backward[from_node].append(to_node)

        return backward

    @lazy_property
    def input_layers(self):
        return find_outputs_in_graph(self.backward_graph)

    @lazy_property
    def output_layers(self):
        return find_outputs_in_graph(self.forward_graph)

    @lazy_property
    def inputs(self):
        placeholders = []

        for layer in self.input_layers:
            placeholder = tf.placeholder(
                tf.float32,
                shape=tf_utils.shape_to_tuple(layer.input_shape),
                name="placeholder/input/{}".format(layer.name),
            )
            placeholders.append(placeholder)

        return make_one_if_possible(placeholders)

    @lazy_property
    def targets(self):
        placeholders = []

        for layer in self.output_layers:
            placeholder = tf.placeholder(
                tf.float32,
                shape=tf_utils.shape_to_tuple(layer.output_shape),
                name="placeholder/target/{}".format(layer.name),
            )
            placeholders.append(placeholder)

        return make_one_if_possible(placeholders)

    @lazy_property
    def outputs(self):
        networks_output = self.output(*as_tuple(self.inputs))
        tf_utils.initialize_uninitialized_variables()
        return networks_output

    @lazy_property
    def training_outputs(self):
        networks_output = self.output(*as_tuple(self.inputs), training=True)
        tf_utils.initialize_uninitialized_variables()
        return networks_output

    def __gt__(self, other):
        left, right = self, other
        self.events.append(('__gt__', join(left, right)))

        graph = LayerGraph()
        previous_operator = None

        for operator, value in reversed(self.events):
            if operator == previous_operator:
                break

            if operator == '__gt__':
                # It's important to put `value` before graph, because
                # we merge in reverse order and we need to make sure
                # that every new value has higher priority.
                graph = merge(value, graph)

            previous_operator = operator

        return graph

    def __bool__(self):
        self.events.append(('__bool__', self))
        return True

    def __nonzero__(self):
        return self.__bool__()  # Hack for python 2

    def __rshift__(self, other):
        return join(self, other)

    def __irshift__(self, other):
        return self.__rshift__(other)

    def __or__(self, other):
        return parallel(self, other)

    def __ior__(self, other):
        return self.__or__(other)

    @abstractmethod
    def output(self, inputs):
        raise NotImplementedError()

    @property
    @abstractmethod
    def output_shape(self):
        raise NotImplementedError()

    @abstractmethod
    def get_output_shape(self, input_shape):
        raise NotImplementedError()


class LayerGraph(BaseGraph):
    def __init__(self, forward_graph=None):
        super(LayerGraph, self).__init__(forward_graph)

        # This allows to run simple check that ensures that
        # created graph have defined layer shape
        self.output_shape

    def clean_layer_references(self, layer_references):
        layers = []

        for layer_reference in layer_references:
            if isinstance(layer_reference, six.string_types):
                layer_reference = self.layer(layer_reference)
            layers.append(layer_reference)

        return layers

    def slice(self, directed_graph, layers):
        layers = self.clean_layer_references(layers)
        forward_graph = self.forward_graph

        if all(layer not in forward_graph for layer in layers):
            unused_layer = next(l for l in layers if l not in forward_graph)
            raise ValueError(
                "Layer `{}` is not used in the graph. Graph: {}, "
                "Layer: {}".format(unused_layer.name, self, unused_layer))

        observed_layers = []
        layers = copy.copy(layers)

        while layers:
            current_layer = layers.pop()
            observed_layers.append(current_layer)

            for next_layer in directed_graph[current_layer]:
                if next_layer not in observed_layers:
                    layers.append(next_layer)

        forward_subgraph = filter_graph(forward_graph, observed_layers)
        return self.__class__(forward_subgraph)

    def end(self, *output_layers):
        return self.slice(self.backward_graph, output_layers)

    def start(self, *input_layers):
        return self.slice(self.forward_graph, input_layers)

    @lazy_property
    def layers(self):
        return list(self)

    def layer(self, layer_name):
        if not isinstance(layer_name, six.string_types):
            raise ValueError(
                "Layer name expected to be a string, "
                "got value {}".format(layer_name))

        layers = []

        for layer in self.forward_graph:
            if layer.name == layer_name:
                layers.append(layer)

        if not layers:
            raise NameError(
                "Cannot find layer with name {!r}".format(layer_name))

        if len(layers) >= 2:
            raise NameError(
                "Ambiguous layer name `{}`. Network has {} "
                "layers with the same name. Layers: {}".format(
                    layer_name, len(layers), layers))

        return layers[0]

    @lazy_property
    def input_shapes(self):
        return [tf.TensorShape(l.input_shape) for l in self.input_layers]

    @lazy_property
    def input_shape(self):
        return make_one_if_possible(self.input_shapes)

    @lazy_property
    def output_shape(self):
        return self.get_output_shape(*self.input_shapes)

    @lazy_property
    def output_shapes_per_layer(self):
        return self.propagate_forward(
            copy.deepcopy(self.input_shapes),
            method='get_output_shape')

    def get_output_shape(self, *inputs):
        outputs = self.propagate_forward(
            copy.deepcopy(inputs),
            method='get_output_shape',
        )
        return make_one_if_possible(
            [outputs[l] for l in self.output_layers])

    def create_variables(self):
        output_shapes = self.output_shapes_per_layer
        backward_graph = self.backward_graph

        for layer in self:
            input_shapes = [layer.input_shape]
            from_layers = backward_graph[layer]

            if layer.frozen:
                continue

            if from_layers:
                input_shapes = [output_shapes[l] for l in from_layers]

            layer.create_variables(*input_shapes)
            layer.frozen = True

    def output(self, *inputs, **kwargs):
        self.create_variables()
        outputs = self.propagate_forward(inputs, method='output', **kwargs)
        return make_one_if_possible([outputs[l] for l in self.output_layers])

    def preformat_inputs(self, inputs):
        if len(inputs) == 1 and isinstance(inputs[0], dict):
            inputs = inputs[0]

        if not isinstance(inputs, dict):
            n_input_layers = len(self.input_layers)
            n_input_vars = len(inputs)

            if n_input_vars != n_input_layers:
                raise ValueError(
                    "Connection has {} input layer(s), but {} inputs was "
                    "provided".format(n_input_layers, n_input_vars))

            inputs = dict(zip(self.input_layers, inputs))

        prepared_inputs = {}
        for layer, input_variable in inputs.items():
            if isinstance(layer, six.string_types):
                layer = self.layer(layer)

            if layer not in self.forward_graph:
                raise ValueError(
                    "The `{}` layer doesn't appear in the network"
                    "".format(layer.name))

            if layer not in self.input_layers:
                raise ValueError(
                    "`{}` is not an input layer in the network"
                    "".format(layer.name))

            prepared_inputs[layer] = input_variable

        return prepared_inputs

    def pass_through_the_layer(self, layer, method, *args, **kwargs):
        layer_method = getattr(layer, method)

        try:
            return layer_method(*args, **kwargs)
        except Exception as exception:
            modified_exception = exception.__class__(
                "{original_message}. Exception occurred while propagating "
                "data through the method `{method}`. Layer: {layer!r}".format(
                    original_message=str(exception).strip('.'),
                    method=method, layer=layer
                )
            )

            if hasattr(sys, 'last_traceback') and six.PY3:
                modified_exception = modified_exception.with_traceback(
                    sys.last_traceback)

            raise modified_exception

    def propagate_forward(self, inputs, method, **kwargs):
        backward_graph = self.backward_graph
        inputs = self.preformat_inputs(inputs)
        outputs = copy.copy(inputs)

        for layer, layer_input in inputs.items():
            outputs[layer] = self.pass_through_the_layer(
                layer, method, layer_input, **kwargs)

        for layer in (l for l in self if l not in inputs):
            layer_inputs = [outputs[l] for l in backward_graph[layer]]
            outputs[layer] = self.pass_through_the_layer(
                layer, method, *layer_inputs, **kwargs)

        return outputs

    @property
    def variables(self):
        self.create_variables()

        variables = OrderedDict()
        observed_variables = []

        for layer in self:
            for name, value in layer.variables.items():
                if value not in observed_variables:
                    observed_variables.append(value)
                    variables[(layer, name)] = value

        return variables

    @property
    def n_parameters(self):
        n_parameters = 0

        for variable in self.variables.values():
            n_parameters += variable.shape.num_elements()

        return n_parameters

    def predict(self, *inputs, **kwargs):
        session = tf_utils.tensorflow_session()

        batch_size = kwargs.pop('batch_size', None)
        verbose = kwargs.pop('verbose', True)

        # We require do to this check for python 2 compatibility
        if kwargs:
            raise TypeError("Unknown arguments: {}".format(kwargs))

        def single_batch_predict(*inputs):
            feed_dict = dict(zip(as_tuple(self.inputs), inputs))
            return session.run(self.outputs, feed_dict=feed_dict)

        outputs = iters.apply_batches(
            function=single_batch_predict,
            inputs=inputs,
            batch_size=batch_size,
            show_progressbar=verbose,
        )
        return np.concatenate(outputs, axis=0)

    def is_sequential(self):
        if len(self.input_layers) > 1 or len(self.output_layers) > 1:
            return False

        forward_graph_layers = self.forward_graph.values()
        backward_graph_layers = self.backward_graph.values()

        for layers in chain(forward_graph_layers, backward_graph_layers):
            if len(layers) >= 2:
                # One of the layers has multiple input
                # or output networks
                return False

        return True

    def layer_names_only(self):
        prepared_graph = OrderedDict()

        for from_layer, to_layers in self.forward_graph.items():
            prepared_graph[from_layer.name] = [l.name for l in to_layers]

        return list(prepared_graph.items())

    def show(self, filepath=None):
        """
        Generates visual representation of the network. Method will create
        PDF that will be rendered and opened automatically. Graph will
        be stored in the ``filepath`` when it's specified.
        """
        if filepath is None:
            filepath = tempfile.mktemp()

        def layer_uid(layer):
            return str(id(layer))

        digraph = graphviz.Digraph()
        shapes_per_layer = self.output_shapes_per_layer

        for layer in self.forward_graph.keys():
            digraph.node(layer_uid(layer), str(layer.name))

        output_id = 1
        for from_layer, to_layers in self.forward_graph.items():
            for to_layer in to_layers:
                digraph.edge(
                    layer_uid(from_layer),
                    layer_uid(to_layer),
                    label=" {}".format(shapes_per_layer[from_layer]))

            if not to_layers:
                output = 'output-{}'.format(output_id)

                digraph.node(output, 'Output #{}'.format(output_id))
                digraph.edge(
                    layer_uid(from_layer), output,
                    label=" {}".format(shapes_per_layer[from_layer]))

                output_id += 1

        digraph.render(filepath, view=True)

    def get_params(self):
        return {'forward_graph': self.forward_graph}

    def __contains__(self, entity):
        return entity in self.forward_graph

    def __len__(self):
        return len(self.forward_graph)

    def __iter__(self):
        for layer in topological_sort(self.backward_graph):
            yield layer

    def __repr__(self):
        if not self.forward_graph:
            return "[empty graph]"

        def format_shapes(shape):
            if isinstance(shape, tf.TensorShape):
                return str(shape)

            shapes = ', '.join([format_shapes(s) for s in shape])
            return '[' + shapes + ']'

        return '{} -> [... {} layers ...] -> {}'.format(
            format_shapes(self.input_shape),
            len(self),
            format_shapes(self.output_shape))


def validate_graphs_before_combining(left_graph, right_graph):
    left_out_layers = left_graph.output_layers
    right_in_layers = right_graph.input_layers

    if len(left_out_layers) > 1 and len(right_in_layers) > 1:
        raise LayerConnectionError(
            "Cannot make many to many connection between graphs. One graph "
            "has {n_left_outputs} outputs (layer names: {left_names}) and "
            "the other one has {n_right_inputs} inputs (layer names: "
            "{right_names}). First graph: {left_graph}, Second graph: "
            "{right_graph}".format(
                left_graph=left_graph,
                n_left_outputs=len(left_out_layers),
                left_names=[layer.name for layer in left_out_layers],

                right_graph=right_graph,
                n_right_inputs=len(right_in_layers),
                right_names=[layer.name for layer in right_in_layers],
            )
        )

    left_out_shapes = as_tuple(left_graph.output_shape)
    right_in_shapes = as_tuple(right_graph.input_shape)

    for left_layer, left_out_shape in zip(left_out_layers, left_out_shapes):
        right = zip(right_in_layers, right_in_shapes)

        for right_layer, right_in_shape in right:
            if left_out_shape.is_compatible_with(right_in_shape):
                continue

            raise LayerConnectionError(
                "Cannot connect layer `{left_name}` to layer `{right_name}`, "
                "because output shape ({left_out_shape}) of the first layer "
                "is incompatible with the input shape ({right_in_shape}) "
                "of the second layer. First layer: {left_layer}, Second "
                "layer: {right_layer}".format(
                    left_layer=left_layer,
                    left_name=left_layer.name,
                    left_out_shape=left_out_shape,

                    right_layer=right_layer,
                    right_name=right_layer.name,
                    right_in_shape=right_in_shape,
                )
            )


def merge(left_graph, right_graph, combine=False):
    """
    Merges two graphs into single one. When ``combine=False`` new
    connection won't be created. And when ``combine=True`` input layers
    from the ``left_graph`` will be combined to the output layers from
    ``right_graph``.

    Parameters
    ----------
    left_graph : layer, network
    right_graph : layer, network

    combine : bool
        Defaults to ``False``.
    """
    if combine:
        validate_graphs_before_combining(left_graph, right_graph)

    forward_graph = OrderedDict()

    for key, value in left_graph.forward_graph.items():
        # To make sure that we copied lists inside of the
        # dictionary, but didn't copied values inside of the list
        forward_graph[key] = copy.copy(value)

    for key, values in right_graph.forward_graph.items():
        if key in forward_graph:
            for value in values:
                if value not in forward_graph[key]:
                    forward_graph[key].append(value)
        else:
            forward_graph[key] = copy.copy(values)

    if combine:
        for left_out_layer in left_graph.output_layers:
            for right_in_layer in right_graph.input_layers:
                forward_graph[left_out_layer].append(right_in_layer)

    if is_cyclic(forward_graph):
        raise LayerConnectionError(
            "Cannot define connection between layers, because it creates "
            "cycle in the graph. Left graph: {}, Right graph: {}"
            "".format(left_graph, right_graph))

    return LayerGraph(forward_graph)


def parallel(*networks):
    """
    Merges all networks/layers into single network without joining
    input and output layers together.

    Parameters
    ----------
    *networks
        Layers or networks. Each network can be specified as a list or
        tuple. In this case, this input will be passed to the ``join``
        function in order to create full network from the specified
        list of the connections.

    Examples
    --------
    >>> from neupy.layers import *
    >>> network = parallel(
    ...     Input(10),
    ...     Input(9) >> Relu(5),
    ...     Relu(5),
    ... )
    [(?, 10), (?, 9), <unknown>] -> [... 4 layers ...] -> \
    [(?, 10), (?, 5), (?, 5)]

    Networks can be specified as a list of layers.

    >>> from neupy.layers import *
    >>> network = parallel([
    ...     Input((28, 28, 1)),
    ...     Convolution((3, 3, 16)),
    ... ], [
    ...     Input((28, 28, 1)),
    ...     Convolution((3, 3, 12)),
    ... ])
    >>>
    >>> network
    [(?, 28, 28, 1), (?, 28, 28, 1)] -> [... 4 layers ...] -> \
    [(?, 26, 26, 16), (?, 26, 26, 12)]
    """
    graph = LayerGraph()

    for network in networks:
        if isinstance(network, (list, tuple)):
            network = join(*network)
        graph = merge(graph, network)

    return graph


def join(*networks):
    """
    Sequentially combines layers and networks into single network.
    Function will go sequentially over layers/networks and
    combine output layers from the one network that it picks
    to the input layers from the next network in the sequence.

    Parameters
    ----------
    *networks
        Layers or networks.

    Examples
    --------
    >>> from neupy.layers import *
    >>> network = join(
    ...     Input((28, 28, 1)),
    ...     Convolution((3, 3, 16)) >> Relu(),
    ...     Convolution((3, 3, 16)) >> Relu(),
    ...     Reshape(),
    ...     Softmax(10),
    ... )
    >>> network
    (?, 28, 28, 1) -> [... 7 layers ...] -> (?, 10)
    """
    graph = LayerGraph()

    for network in networks:
        graph = merge(graph, network, combine=True)

    return graph


def repeat(network_or_layer, n):
    """
    Function copies input `n - 1` times and connects everything in sequential
    order.

    Parameters
    ----------
    network_or_layer : network or layer
        Layer or network (connection of layers).

    n : int
        Number of times input should be replicated.

    Examples
    --------
    >>> from neupy.layers import *
    >>>
    >>> block = Conv((3, 3, 32)) >> Relu() >> BN()
    >>> block
    <unknown> -> [... 3 layers ...] -> (?, ?, ?, 32)
    >>>
    >>> repeat(block, n=5)
    <unknown> -> [... 15 layers ...] -> (?, ?, ?, 32)
    """
    if n <= 0 or not isinstance(n, int):
        raise ValueError(
            "The `n` parameter should be a positive integer, "
            "got {} instead".format(n))

    if n == 1:
        return network_or_layer

    input_shape = network_or_layer.input_shape
    output_shape = network_or_layer.output_shape

    if not input_shape.is_compatible_with(output_shape):
        raise LayerConnectionError(
            "Cannot connect network/layer to its copy, because input "
            "shape is incompatible with the output shape. Input shape: {}, "
            "Output shape: {}".format(input_shape, output_shape))

    new_networks = [copy.deepcopy(network_or_layer) for _ in range(n - 1)]
    return join(network_or_layer, *new_networks)



import types

import tensorflow as tf

from neupy import init
from neupy.utils import tensorflow_session
from neupy.exceptions import LayerConnectionError
from neupy.core.properties import (
    IntProperty, Property,
    NumberProperty, ParameterProperty,
)
from .base import BaseLayer


__all__ = ('LSTM', 'GRU')


def clip_gradient(value, clip_value):
    if not hasattr(clip_gradient, 'added_gradients'):
        clip_gradient.added_gradients = set()

    session = tensorflow_session()
    graph = session.graph
    operation_name = "ClipGradient-" + str(clip_value)

    if operation_name not in clip_gradient.added_gradients:
        # Make sure that we won't create the same operation twice.
        # Otherwise tensorflow will trigger an exception.
        @tf.RegisterGradient(operation_name)
        def clip_gradient_grad(op, grad):
            return tf.clip_by_value(grad, -clip_value, clip_value)

        clip_gradient.added_gradients.add(operation_name)

    with graph.gradient_override_map({"Identity": operation_name}):
        return tf.identity(value)


def unroll_scan(fn, sequence, outputs_info):
    """
    Helper function to unroll for loops. Can be used to unroll
    ``tensorflow.scan``.

    Parameters
    ----------
    fn : function
        Function that defines calculations at each step.

    sequences : TensorVariable or list of TensorVariables
        List of TensorVariable with sequence data. The function iterates
        over the first dimension of each TensorVariable.

    outputs_info : list of TensorVariables
        List of tensors specifying the initial values for each recurrent
        value.

    Returns
    -------
    List of TensorVariables. Each element in the list gives
    the recurrent values at each time step.
    """
    with tf.name_scope('unroll-scan'):
        outputs = []
        prev_vals = outputs_info

        for entity in tf.unstack(sequence):
            output = fn(prev_vals, entity)
            outputs.append(output[-1])
            prev_vals = output

        return tf.stack(outputs)


class BaseRNNLayer(BaseLayer):
    """
    Base class for the recurrent layers

    Parameters
    ----------
    n_units : int
        Number of hidden units in the layer.

    only_return_final : bool
        If ``True``, only return the final sequential output
        (e.g. for tasks where a single target value for the entire
        sequence is desired). In this case, Tensorfow makes an
        optimization which saves memory. Defaults to ``True``.

    {BaseLayer.name}
    """
    n_units = IntProperty(minval=1)
    only_return_final = Property(expected_type=bool)

    def __init__(self, n_units, only_return_final=True, name=None):
        super(BaseRNNLayer, self).__init__(name=name)
        self.only_return_final = only_return_final
        self.n_units = n_units

    def fail_if_shape_invalid(self, input_shape):
        if input_shape and input_shape.ndims != 3:
            clsname = self.__class__.__name__
            raise LayerConnectionError(
                "{} layer was expected input with three dimensions, "
                "but got input with {} dimensions instead. Layer: {}"
                "".format(clsname, input_shape.ndims, self))

    def get_output_shape(self, input_shape):
        input_shape = tf.TensorShape(input_shape)
        n_samples = input_shape[0]

        self.fail_if_shape_invalid(input_shape)

        if self.only_return_final:
            return tf.TensorShape((n_samples, self.n_units))

        n_time_steps = input_shape[1]
        return tf.TensorShape((n_samples, n_time_steps, self.n_units))


class LSTM(BaseRNNLayer):
    """
    Long Short Term Memory (LSTM) Layer.

    Parameters
    ----------
    {BaseRNNLayer.n_units}

    {BaseRNNLayer.only_return_final}

    input_weights : Initializer, ndarray
        Weight parameters for input connection.
        Defaults to :class:`HeNormal() <neupy.init.HeNormal>`.

    hidden_weights : Initializer, ndarray
        Weight parameters for hidden connection.
        Defaults to :class:`HeNormal() <neupy.init.HeNormal>`.

    cell_weights : Initializer, ndarray
        Weight parameters for cell connection. Require only when
        ``peepholes=True`` otherwise it will be ignored.
        Defaults to :class:`HeNormal() <neupy.init.HeNormal>`.

    bias : Initializer, ndarray
        Bias parameters for all gates.
        Defaults to :class:`Constant(0) <neupy.init.Constant>`.

    ingate : function
        Activation function for the input gate.
        Defaults to ``tf.nn.sigmoid``.

    forgetgate : function
        Activation function for the forget gate.
        Defaults to ``tf.nn.sigmoid``.

    outgate : function
        Activation function for the output gate.
        Defaults to ``tf.nn.sigmoid``.

    cell : function
        Activation function for the cell.
        Defaults to ``tf.tanh``.

    learn_init : bool
        If ``True``, make ``cell_init`` and ``hidden_init`` trainable
        variables. Defaults to ``False``.

    cell_init : array-like, Tensorfow variable, scalar or Initializer
        Initializer for initial cell state (:math:`c_0`).
        Defaults to :class:`Constant(0) <neupy.init.Constant>`.

    hidden_init : array-like, Tensorfow variable, scalar or Initializer
        Initializer for initial hidden state (:math:`h_0`).
        Defaults to :class:`Constant(0) <neupy.init.Constant>`.

    backwards : bool
        If ``True``, process the sequence backwards and then reverse the
        output again such that the output from the layer is always
        from :math:`x_1` to :math:`x_n`. Defaults to ``False``

    peepholes : bool
        If ``True``, the LSTM uses peephole connections.
        When ``False``, cell parameters  are ignored.
        Defaults to ``False``.

    unroll_scan : bool
        If ``True`` the recursion is unrolled instead of using scan.
        For some graphs this gives a significant speed up but it
        might also consume more memory. When ``unroll_scan=True``,
        backpropagation always includes the full sequence, so
        ``n_gradient_steps`` must be set to ``-1`` and the input
        sequence length must be known at compile time (i.e.,
        cannot be given as ``None``). Defaults to ``False``.

    gradient_clipping : float or int
        If nonzero, the gradient messages are clipped to the
        given value during the backward pass. Defaults to ``0``.

    {BaseLayer.name}

    Notes
    -----
    Code was adapted from the
    `Lasagne <https://github.com/Lasagne/Lasagne>`_ library.

    Examples
    --------
    Sequence classification

    >>> from neupy.layers import *
    >>>
    >>> n_time_steps = 40
    >>> n_categories = 20
    >>> embedded_size = 10
    >>>
    >>> network = join(
    ...     Input(n_time_steps),
    ...     Embedding(n_categories, embedded_size),
    ...     LSTM(20),
    ...     Sigmoid(1),
    ... )
    >>> network
    (?, 40) -> [... 4 layers ...] -> (?, 1)
    """
    input_weights = ParameterProperty()
    hidden_weights = ParameterProperty()
    cell_weights = ParameterProperty()
    biases = ParameterProperty()

    ingate = Property(expected_type=types.FunctionType)
    forgetgate = Property(expected_type=types.FunctionType)
    outgate = Property(expected_type=types.FunctionType)
    cell = Property(expected_type=types.FunctionType)

    learn_init = Property(expected_type=bool)
    cell_init = ParameterProperty()
    hidden_init = ParameterProperty()

    unroll_scan = Property(expected_type=bool)
    backwards = Property(expected_type=bool)
    peepholes = Property(expected_type=bool)
    gradient_clipping = NumberProperty(minval=0)

    def __init__(self, n_units, only_return_final=True,
                 # Trainable parameters
                 input_weights=init.HeNormal(),
                 hidden_weights=init.HeNormal(),
                 cell_weights=init.HeNormal(), biases=0,
                 # Activation functions
                 ingate=tf.nn.sigmoid, forgetgate=tf.nn.sigmoid,
                 outgate=tf.nn.sigmoid, cell=tf.tanh,
                 # Cell states
                 cell_init=0, hidden_init=0, learn_init=False,
                 # Misc
                 unroll_scan=False, backwards=False, peepholes=False,
                 gradient_clipping=0, name=None):

        super(LSTM, self).__init__(
            n_units=n_units,
            only_return_final=only_return_final,
            name=name,
        )

        self.input_weights = input_weights
        self.hidden_weights = hidden_weights
        self.cell_weights = cell_weights
        self.biases = biases

        self.ingate = ingate
        self.forgetgate = forgetgate
        self.outgate = outgate
        self.cell = cell

        self.learn_init = learn_init
        self.cell_init = cell_init
        self.hidden_init = hidden_init

        self.unroll_scan = unroll_scan
        self.backwards = backwards
        self.peepholes = peepholes
        self.gradient_clipping = gradient_clipping

    def create_variables(self, input_shape):
        self.input_shape = input_shape
        self.input_weights = self.variable(
            value=self.input_weights,
            name='input_weights',
            shape=(input_shape[-1], 4 * self.n_units),
        )
        self.hidden_weights = self.variable(
            value=self.hidden_weights, name='hidden_weights',
            shape=(self.n_units, 4 * self.n_units),
        )
        self.biases = self.variable(
            value=self.biases, name='biases',
            shape=(4 * self.n_units,),
        )
        self.cell_init = self.variable(
            value=self.cell_init,
            shape=(1, self.n_units),
            name="cell_init",
            trainable=self.learn_init,
        )
        self.hidden_init = self.variable(
            value=self.hidden_init,
            shape=(1, self.n_units),
            name="hidden_init",
            trainable=self.learn_init,
        )

        # If peephole (cell to gate) connections were enabled, initialize
        # peephole connections.  These are elementwise products with the cell
        # state, so they are represented as vectors.
        if self.peepholes:
            self.weight_cell_to_ingate = self.variable(
                value=self.cell_weights,
                name='weight_cell_to_ingate',
                shape=(self.n_units,))
            self.weight_cell_to_forgetgate = self.variable(
                value=self.cell_weights,
                name='weight_cell_to_forgetgate',
                shape=(self.n_units,))
            self.weight_cell_to_outgate = self.variable(
                value=self.cell_weights,
                name='weight_cell_to_outgate',
                shape=(self.n_units,))

    def output(self, input, **kwargs):
        # Because scan iterates over the first dimension we
        # dimshuffle to (n_time_steps, n_samples, n_features)
        input = tf.transpose(input, [1, 0, 2])

        def one_lstm_step(states, input_n):
            with tf.name_scope('lstm-cell'):
                cell_previous, hid_previous = states
                input_n = tf.matmul(input_n, self.input_weights) + self.biases

                # Calculate gates pre-activations and slice
                gates = input_n + tf.matmul(hid_previous, self.hidden_weights)

                # Clip gradients
                if self.gradient_clipping != 0:
                    gates = clip_gradient(gates, self.gradient_clipping)

                # Extract the pre-activation gate values
                ingate, forgetgate, cell_input, outgate = tf.split(
                    gates, 4, axis=1)

                if self.peepholes:
                    # Compute peephole connections
                    ingate += cell_previous * self.weight_cell_to_ingate
                    forgetgate += (
                        cell_previous * self.weight_cell_to_forgetgate)

                # Apply nonlinearities
                ingate = self.ingate(ingate)
                forgetgate = self.forgetgate(forgetgate)
                cell_input = self.cell(cell_input)

                # Compute new cell value
                cell = forgetgate * cell_previous + ingate * cell_input

                if self.peepholes:
                    outgate += cell * self.weight_cell_to_outgate

                outgate = self.outgate(outgate)

                # Compute new hidden unit activation
                hid = outgate * tf.tanh(cell)
                return [cell, hid]

        input_shape = tf.shape(input)
        n_samples = input_shape[1]  # batch dim has been moved
        cell_init = tf.tile(self.cell_init, (n_samples, 1))
        hidden_init = tf.tile(self.hidden_init, (n_samples, 1))
        sequence = input

        if self.backwards:
            sequence = tf.reverse(sequence, axis=[0])

        if self.unroll_scan:
            # Explicitly unroll the recurrence instead of using scan
            hid_out = unroll_scan(
                fn=one_lstm_step,
                sequence=sequence,
                outputs_info=[cell_init, hidden_init],
            )
        else:
            _, hid_out = tf.scan(
                fn=one_lstm_step,
                elems=sequence,
                initializer=[cell_init, hidden_init],
                name='lstm-scan',
            )

        # When it is requested that we only return the final sequence step,
        # we need to slice it out immediately after scan is applied
        if self.only_return_final:
            return hid_out[-1]

        # if scan is backward reverse the output
        if self.backwards:
            hid_out = tf.reverse(hid_out, axis=[0])

        # dimshuffle back to (n_samples, n_time_steps, n_features))
        hid_out = tf.transpose(hid_out, [1, 0, 2])

        return hid_out


class GRU(BaseRNNLayer):
    """
    Gated Recurrent Unit (GRU) Layer.

    Parameters
    ----------
    {BaseRNNLayer.n_units}

    {BaseRNNLayer.only_return_final}

    input_weights : Initializer, ndarray
        Weight parameters for input connection.
        Defaults to :class:`HeNormal() <neupy.init.HeNormal>`.

    hidden_weights : Initializer, ndarray
        Weight parameters for hidden connection.
        Defaults to :class:`HeNormal() <neupy.init.HeNormal>`.

    biases : Initializer, ndarray
        Bias parameters for all gates.
        Defaults to :class:`Constant(0) <neupy.init.Constant>`.

    resetgate : function
        Activation function for the reset gate.
        Defaults to ``tf.nn.sigmoid``.

    updategate : function
        Activation function for the update gate.
        Defaults to ``tf.nn.sigmoid``.

    hidden_update : function
        Activation function for the hidden state update.
        Defaults to ``tf.tanh``.

    learn_init : bool
        If ``True``, make ``hidden_init`` trainable variable.
        Defaults to ``False``.

    hidden_init : array-like, Tensorfow variable, scalar or Initializer
        Initializer for initial hidden state (:math:`h_0`).
        Defaults to :class:`Constant(0) <neupy.init.Constant>`.

    backwards : bool
        If ``True``, process the sequence backwards and then reverse the
        output again such that the output from the layer is always
        from :math:`x_1` to :math:`x_n`. Defaults to ``False``.

    unroll_scan : bool
        If ``True`` the recursion is unrolled instead of using scan.
        For some graphs this gives a significant speed up but it
        might also consume more memory. When ``unroll_scan=True``,
        backpropagation always includes the full sequence, so
        ``n_gradient_steps`` must be set to ``-1`` and the input
        sequence length must be known at compile time (i.e.,
        cannot be given as ``None``). Defaults to ``False``.

    {BaseLayer.name}

    Notes
    -----
    Code was adapted from the
    `Lasagne <https://github.com/Lasagne/Lasagne>`_ library.

    Examples
    --------
    Sequence classification

    >>> from neupy.layers import *
    >>>
    >>> n_time_steps = 40
    >>> n_categories = 20
    >>> embedded_size = 10
    >>>
    >>> network = join(
    ...     Input(n_time_steps),
    ...     Embedding(n_categories, embedded_size),
    ...     GRU(20),
    ...     Sigmoid(1),
    ... )
    >>> network
    (?, 40) -> [... 4 layers ...] -> (?, 1)
    """
    input_weights = ParameterProperty()
    hidden_weights = ParameterProperty()
    biases = ParameterProperty()

    resetgate = Property(expected_type=types.FunctionType)
    updategate = Property(expected_type=types.FunctionType)
    hidden_update = Property(expected_type=types.FunctionType)

    hidden_init = ParameterProperty()
    learn_init = Property(default=False, expected_type=bool)

    backwards = Property(expected_type=bool)
    unroll_scan = Property(expected_type=bool)
    gradient_clipping = NumberProperty(default=0, minval=0)

    def __init__(self, n_units, only_return_final=True,
                 # Trainable parameters
                 input_weights=init.HeNormal(),
                 hidden_weights=init.HeNormal(),
                 biases=0,
                 # Activation functions
                 resetgate=tf.nn.sigmoid,
                 updategate=tf.nn.sigmoid,
                 hidden_update=tf.tanh,
                 # Cell states
                 hidden_init=0, learn_init=False,
                 # Misc
                 unroll_scan=False, backwards=False,
                 gradient_clipping=0, name=None):

        super(GRU, self).__init__(
            n_units=n_units,
            only_return_final=only_return_final,
            name=name,
        )

        self.input_weights = input_weights
        self.hidden_weights = hidden_weights
        self.biases = biases

        self.resetgate = resetgate
        self.updategate = updategate
        self.hidden_update = hidden_update

        self.hidden_init = hidden_init
        self.learn_init = learn_init

        self.unroll_scan = unroll_scan
        self.backwards = backwards
        self.gradient_clipping = gradient_clipping

    def create_variables(self, input_shape):
        self.input_weights = self.variable(
            value=self.input_weights,
            name='input_weights',
            shape=(input_shape[-1], 3 * self.n_units),
        )
        self.hidden_weights = self.variable(
            value=self.hidden_weights,
            name='hidden_weights',
            shape=(self.n_units, 3 * self.n_units),
        )
        self.biases = self.variable(
            value=self.biases, name='biases',
            shape=(3 * self.n_units,),
        )
        self.hidden_init = self.variable(
            value=self.hidden_init,
            shape=(1, self.n_units),
            name="hidden_init",
            trainable=self.learn_init
        )

    def output(self, input, **kwargs):
        # Because scan iterates over the first dimension we
        # dimshuffle to (n_time_steps, n_samples, n_features)
        input = tf.transpose(input, [1, 0, 2])

        # Create single recurrent computation step function
        # input_n is the n'th vector of the input
        def one_gru_step(states, input_n):
            with tf.name_scope('gru-cell'):
                hid_previous, = states
                input_n = tf.matmul(input_n, self.input_weights) + self.biases

                # Compute W_{hr} h_{t - 1}, W_{hu} h_{t - 1},
                # and W_{hc} h_{t - 1}
                hid_input = tf.matmul(hid_previous, self.hidden_weights)

                if self.gradient_clipping != 0:
                    input_n = clip_gradient(input_n, self.gradient_clipping)
                    hid_input = clip_gradient(
                        hid_input, self.gradient_clipping)

                hid_resetgate, hid_updategate, hid_hidden = tf.split(
                    hid_input, 3, axis=1)

                in_resetgate, in_updategate, in_hidden = tf.split(
                    input_n, 3, axis=1)

                # Reset and update gates
                resetgate = self.resetgate(hid_resetgate + in_resetgate)
                updategate = self.updategate(hid_updategate + in_updategate)

                # Compute W_{xc}x_t + r_t \odot (W_{hc} h_{t - 1})
                hidden_update = in_hidden + resetgate * hid_hidden

                if self.gradient_clipping != 0:
                    hidden_update = clip_gradient(
                        hidden_update, self.gradient_clipping)

                hidden_update = self.hidden_update(hidden_update)

                # Compute (1 - u_t)h_{t - 1} + u_t c_t
                return [
                    hid_previous - updategate * (hid_previous - hidden_update)
                ]

        input_shape = tf.shape(input)
        n_samples = input_shape[1]  # batch dim has been moved
        hidden_init = tf.tile(self.hidden_init, (n_samples, 1))
        sequence = input

        if self.backwards:
            sequence = tf.reverse(sequence, axis=[0])

        if self.unroll_scan:
            # Explicitly unroll the recurrence instead of using scan
            hid_out = unroll_scan(
                fn=one_gru_step,
                sequence=sequence,
                outputs_info=[hidden_init]
            )
        else:
            hid_out, = tf.scan(
                fn=one_gru_step,
                elems=sequence,
                initializer=[hidden_init],
                name='gru-scan',
            )

        # When it is requested that we only return the final sequence step,
        # we need to slice it out immediately after scan is applied
        if self.only_return_final:
            return hid_out[-1]

        # if scan is backward reverse the output
        if self.backwards:
            hid_out = tf.reverse(hid_out, axis=[0])

        # dimshuffle back to (n_samples, n_time_steps, n_features))
        hid_out = tf.transpose(hid_out, [1, 0, 2])
        return hid_out



# -*- coding: utf-8 -*-
from __future__ import division, absolute_import, unicode_literals

import time
import inspect
from abc import abstractmethod

import numpy as np

from neupy.exceptions import StopTraining
from neupy.core.logs import Verbose
from neupy.core.config import ConfigurableABC
from neupy.core.properties import Property, NumberProperty, IntProperty
from neupy.algorithms import signals as base_signals
from neupy.algorithms.plots import plot_optimizer_errors
from neupy.utils import iters, as_tuple


__all__ = ('BaseSkeleton', 'BaseNetwork')


def preformat_value(value):
    if inspect.isfunction(value) or inspect.isclass(value):
        return value.__name__

    elif isinstance(value, (list, tuple, set)):
        return [preformat_value(v) for v in value]

    elif isinstance(value, (np.ndarray, np.matrix)):
        return value.shape

    return value


class BaseSkeleton(ConfigurableABC, Verbose):
    """
    Base class for neural network algorithms.

    Methods
    -------
    fit(\*args, \*\*kwargs)
        Alias to the ``train`` method.

    predict(X)
        Predicts output for the specified input.
    """

    def __init__(self, *args, **options):
        super(BaseSkeleton, self).__init__(*args, **options)

        self.logs.title("Main information")
        self.logs.message("ALGORITHM", self.__class__.__name__)
        self.logs.newline()

        for key, data in sorted(self.options.items()):
            formated_value = preformat_value(getattr(self, key))
            msg_text = "{} = {}".format(key, formated_value)
            self.logs.message("OPTION", msg_text, color='green')

        self.logs.newline()

    @abstractmethod
    def train(self, X, y):
        raise NotImplementedError()

    @abstractmethod
    def predict(self, X):
        raise NotImplementedError()

    def transform(self, X):
        return self.predict(X)

    def fit(self, X, y=None, *args, **kwargs):
        self.train(X, y, *args, **kwargs)
        return self

    def repr_options(self):
        options = []
        for option_name in self.options:
            option_value = getattr(self, option_name)
            option_value = preformat_value(option_value)

            option_repr = "{}={}".format(option_name, option_value)
            options.append(option_repr)

        return ', '.join(options)

    def __repr__(self):
        class_name = self.__class__.__name__
        available_options = self.repr_options()
        return "{}({})".format(class_name, available_options)

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__.update(state)


class Events(object):
    def __init__(self, network, signals):
        self.network = network
        self.signals = signals
        self.logs = []

    def trigger(self, name, store_data=False, **data):
        if store_data and data:
            self.logs.append(dict(data, name=name))

        for signal in self.signals:
            if hasattr(signal, name):
                signal_method = getattr(signal, name)
                signal_method(self.network, **data)


class BaseNetwork(BaseSkeleton):
    """
    Base class for Neural Network algorithms.

    Parameters
    ----------
    step : float
        Learning rate, defaults to ``0.1``.

    show_epoch : int
        This property controls how often the network will display
        information about training. It has to be defined as positive
        integer. For instance, number ``100`` mean that network shows
        summary at 1st, 100th, 200th, 300th ... and last epochs.

        Defaults to ``1``.

    shuffle_data : bool
        If it's ``True`` than training data will be shuffled before
        the training. Defaults to ``True``.

    signals : dict, list or function
        Function that will be triggered after certain events during
        the training.

    {Verbose.Parameters}

    Methods
    -------
    {BaseSkeleton.fit}

    predict(X)
        Propagates input ``X`` through the network and
        returns produced output.

    plot_errors(logx=False, show=True, **figkwargs)
        Using errors collected during the training this method
        generates plot that can give additional insight into the
        performance reached during the training.

    Attributes
    ----------
    errors : list
        Information about errors. It has two main attributes, namely
        ``train`` and ``valid``. These attributes provide access to
        the training and validation errors respectively.

    last_epoch : int
        Value equals to the last trained epoch. After initialization
        it is equal to ``0``.

    n_updates_made : int
        Number of training updates applied to the network.
    """
    step = NumberProperty(default=0.1, minval=0)
    show_epoch = IntProperty(minval=1, default=1)
    shuffle_data = Property(default=False, expected_type=bool)
    signals = Property(expected_type=object)

    def __init__(self, *args, **options):
        super(BaseNetwork, self).__init__(*args, **options)

        self.last_epoch = 0
        self.n_updates_made = 0
        self.errors = base_signals.ErrorCollector()

        signals = list(as_tuple(
            base_signals.ProgressbarSignal(),
            base_signals.PrintLastErrorSignal(),
            self.errors,
            self.signals,
        ))

        for i, signal in enumerate(signals):
            if inspect.isfunction(signal):
                signals[i] = base_signals.EpochEndSignal(signal)

            elif inspect.isclass(signal):
                signals[i] = signal()

        self.events = Events(network=self, signals=signals)

    def one_training_update(self, X_train, y_train=None):
        """
        Function would be trigger before run all training procedure
        related to the current epoch.

        Parameters
        ----------
        epoch : int
            Current epoch number.
        """
        raise NotImplementedError()

    def score(self, X, y):
        raise NotImplementedError()

    def plot_errors(self, logx=False, show=True, **figkwargs):
        return plot_optimizer_errors(
            optimizer=self,
            logx=logx,
            show=show,
            **figkwargs
        )

    def train(self, X_train, y_train=None, X_test=None, y_test=None,
              epochs=100, batch_size=None):
        """
        Method train neural network.

        Parameters
        ----------
        X_train : array-like
        y_train : array-like or None
        X_test : array-like or None
        y_test : array-like or None

        epochs : int
            Defaults to ``100``.

        epsilon : float or None
            Defaults to ``None``.
        """
        if epochs <= 0:
            raise ValueError("Number of epochs needs to be a positive number")

        epochs = int(epochs)
        first_epoch = self.last_epoch + 1
        batch_size = batch_size or getattr(self, 'batch_size', None)

        self.events.trigger(
            name='train_start',
            X_train=X_train,
            y_train=y_train,
            epochs=epochs,
            batch_size=batch_size,
            store_data=False,
        )

        try:
            for epoch in range(first_epoch, first_epoch + epochs):
                self.events.trigger('epoch_start')

                self.last_epoch = epoch
                iterator = iters.minibatches(
                    (X_train, y_train),
                    batch_size,
                    self.shuffle_data,
                )

                for X_batch, y_batch in iterator:
                    self.events.trigger('update_start')
                    update_start_time = time.time()

                    train_error = self.one_training_update(X_batch, y_batch)
                    self.n_updates_made += 1

                    self.events.trigger(
                        name='train_error',
                        value=train_error,
                        eta=time.time() - update_start_time,
                        epoch=epoch,
                        n_updates=self.n_updates_made,
                        n_samples=iters.count_samples(X_batch),
                        store_data=True,
                    )
                    self.events.trigger('update_end')

                if X_test is not None:
                    test_start_time = time.time()
                    validation_error = self.score(X_test, y_test)
                    self.events.trigger(
                        name='valid_error',
                        value=validation_error,
                        eta=time.time() - test_start_time,
                        epoch=epoch,
                        n_updates=self.n_updates_made,
                        n_samples=iters.count_samples(X_test),
                        store_data=True,
                    )

                self.events.trigger('epoch_end')

        except StopTraining as err:
            self.logs.message(
                "TRAIN",
                "Epoch #{} was stopped. Message: {}".format(epoch, str(err)))

        self.events.trigger('train_end')



import numpy as np
from sklearn import datasets
import matplotlib.pyplot as plt
from neupy import algorithms, utils


plt.style.use('ggplot')
utils.reproducible()


if __name__ == '__main__':
    ggplot_colors = plt.rcParams['axes.prop_cycle']
    colors = np.array([c['color'] for c in ggplot_colors])

    dataset = datasets.load_iris()
    # use only two features in order
    # to make visualization simpler
    data = dataset.data[:, [2, 3]]
    target = dataset.target

    sofm = algorithms.SOFM(
        # Use only two features for the input
        n_inputs=2,

        # Number of outputs defines number of features
        # in the SOFM or in terms of clustering - number
        # of clusters
        n_outputs=20,

        # In clustering application we will prefer that
        # clusters will be updated independently from each
        # other. For this reason we set up learning radius
        # equal to zero
        learning_radius=0,

        # Training step size or learning rate
        step=0.25,

        # Shuffles dataset before every training epoch.
        shuffle_data=True,

        # Instead of generating random weights
        # (features / cluster centers) SOFM will sample
        # them from the data. Which means that after
        # initialization step 3 random data samples will
        # become cluster centers
        weight='sample_from_data',

        # Shows training progress in terminal
        verbose=True,
    )
    sofm.train(data, epochs=200)

    plt.title('Clustering iris dataset with SOFM')
    plt.xlabel('Feature #3')
    plt.ylabel('Feature #4')

    plt.scatter(*data.T, c=colors[target], s=100, alpha=1)
    cluster_centers = plt.scatter(*sofm.weight, s=300, c=colors[3])

    plt.legend([cluster_centers], ['Cluster center'], loc='upper left')
    plt.show()



import matplotlib.pyplot as plt
from neupy import algorithms, utils

from helpers import plot_2d_grid, make_circle


plt.style.use('ggplot')
utils.reproducible()


if __name__ == '__main__':
    GRID_WIDTH = 10
    GRID_HEIGHT = 10

    configurations = [{
        'grid_type': 'hexagon',
        'use_hexagon_grid': True,
        'title': 'Using hexagon grid',
    }, {
        'grid_type': 'rect',
        'use_hexagon_grid': False,
        'title': 'Using regcangular grid',
    }]

    data = make_circle()

    red, blue = ('#E24A33', '#348ABD')
    n_columns = len(configurations)

    plt.figure(figsize=(12, 5))

    for index, conf in enumerate(configurations, start=1):
        sofm = algorithms.SOFM(
            n_inputs=2,
            features_grid=(GRID_HEIGHT, GRID_WIDTH),

            verbose=True,
            shuffle_data=True,
            grid_type=conf['grid_type'],

            learning_radius=8,
            reduce_radius_after=5,

            std=2,
            reduce_std_after=5,

            step=0.3,
            reduce_step_after=5,
        )
        sofm.train(data, epochs=40)

        plt.subplot(1, n_columns, index)

        plt.title(conf['title'])
        plt.scatter(*data.T, color=blue, alpha=0.05)
        plt.scatter(*sofm.weight, color=red)

        weights = sofm.weight.reshape((2, GRID_HEIGHT, GRID_WIDTH))
        plot_2d_grid(weights, color=red, hexagon=conf['use_hexagon_grid'])

    plt.show()



import tensorflow as tf

from neupy.core.properties import BoundedProperty, WithdrawProperty
from neupy.utils import (
    asfloat, flatten, function_name_scope,
    make_single_vector,
)
from neupy.utils.tf_utils import setup_parameter_updates
from .base import BaseOptimizer


__all__ = ('Hessian',)


@function_name_scope
def find_hessian_and_gradient(error_function, parameters):
    """
    Compute hessian matrix and gradient vector.

    Parameters
    ----------
    error_function : Tensor

    parameters : list of Tensorfow variable
        Neural network parameters (e.g. weights, biases).

    Returns
    -------
    Tensorfow variable
    """
    gradients = tf.gradients(error_function, parameters)
    full_gradient = make_single_vector(gradients)

    full_gradient_shape = tf.shape(full_gradient)
    n_samples = full_gradient_shape[0]

    def compute_gradient_per_value(index, result):
        gradients = tf.gradients(full_gradient[index], parameters)
        hessian = make_single_vector(gradients)
        return (index + 1, result.write(index, hessian))

    _, hessian = tf.while_loop(
        lambda index, _: index < n_samples,
        compute_gradient_per_value,
        [
            tf.constant(0, tf.int32),
            tf.TensorArray(tf.float32, size=n_samples),
        ]
    )

    return hessian.stack(), full_gradient


class Hessian(BaseOptimizer):
    """
    Hessian gradient decent optimization, also known as Newton's method. This
    algorithm uses second-order derivative (hessian matrix) in order to
    choose correct step during the training iteration. Because of this,
    method doesn't have ``step`` parameter.

    Parameters
    ----------
    penalty_const : float
        Inverse hessian could be singular matrix. For this reason
        algorithm include penalty that add to hessian matrix identity
        multiplied by defined constant. Defaults to ``1``.

    {BaseOptimizer.network}

    {BaseOptimizer.loss}

    {BaseOptimizer.regularizer}

    {BaseOptimizer.show_epoch}

    {BaseOptimizer.shuffle_data}

    {BaseOptimizer.signals}

    {BaseOptimizer.verbose}

    Attributes
    ----------
    {BaseOptimizer.Attributes}

    Methods
    -------
    {BaseOptimizer.Methods}

    Notes
    -----
    - Method requires all training data during propagation, which means
      it cannot be trained with mini-batches.

    - This method calculates full hessian matrix which means it will compute
      matrix with NxN parameters, where N = number of parameters in the
      network.

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>> from neupy.layers import *
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> network = Input(2) >> Sigmoid(3) >> Sigmoid(1)
    >>> optimizer = algorithms.Hessian(network)
    >>> optimizer.train(x_train, y_train)

    See Also
    --------
    :network:`HessianDiagonal` : Hessian diagonal approximation.
    """
    penalty_const = BoundedProperty(default=1, minval=0)
    step = WithdrawProperty()

    def init_train_updates(self):
        penalty_const = asfloat(self.penalty_const)

        n_parameters = self.network.n_parameters
        variables = self.network.variables
        parameters = [var for var in variables.values() if var.trainable]
        param_vector = make_single_vector(parameters)

        hessian_matrix, full_gradient = find_hessian_and_gradient(
            self.variables.loss, parameters
        )
        parameter_update = tf.matrix_solve(
            hessian_matrix + penalty_const * tf.eye(n_parameters),
            tf.reshape(full_gradient, [-1, 1])
        )
        updated_parameters = param_vector - flatten(parameter_update)
        updates = setup_parameter_updates(parameters, updated_parameters)

        return updates



import numpy as np


__all__ = ('make_digits', 'load_digits')


digits_data = np.array([
    [
        # zero
        0, 1, 1, 0,
        1, 0, 0, 1,
        1, 0, 0, 1,
        1, 0, 0, 1,
        1, 0, 0, 1,
        0, 1, 1, 0,
    ], [
        # one
        0, 1, 1, 0,
        0, 0, 1, 0,
        0, 0, 1, 0,
        0, 0, 1, 0,
        0, 0, 1, 0,
        0, 0, 1, 0,
    ], [
        # two
        1, 1, 1, 0,
        0, 0, 0, 1,
        0, 0, 0, 1,
        0, 1, 1, 0,
        1, 0, 0, 0,
        1, 1, 1, 1,
    ], [
        # three
        1, 1, 1, 0,
        0, 0, 0, 1,
        0, 1, 1, 0,
        0, 0, 0, 1,
        0, 0, 0, 1,
        1, 1, 1, 0,
    ], [
        # four
        0, 0, 1, 1,
        0, 1, 0, 1,
        1, 0, 0, 1,
        1, 1, 1, 1,
        0, 0, 0, 1,
        0, 0, 0, 1,
    ], [
        # five
        0, 1, 1, 1,
        1, 0, 0, 0,
        0, 1, 1, 0,
        0, 0, 0, 1,
        0, 0, 0, 1,
        1, 1, 1, 0,
    ], [
        # six
        0, 1, 1, 0,
        1, 0, 0, 0,
        1, 1, 1, 0,
        1, 0, 0, 1,
        1, 0, 0, 1,
        0, 1, 1, 0,
    ], [
        # seven
        0, 1, 1, 1,
        0, 0, 0, 1,
        0, 0, 0, 1,
        0, 0, 1, 0,
        0, 1, 0, 0,
        0, 1, 0, 0,
    ], [
        # eight
        0, 1, 1, 0,
        1, 0, 0, 1,
        0, 1, 1, 0,
        1, 0, 0, 1,
        1, 0, 0, 1,
        0, 1, 1, 0,
    ], [
        # nine
        0, 1, 1, 0,
        1, 0, 0, 1,
        1, 0, 0, 1,
        0, 1, 1, 1,
        0, 0, 0, 1,
        0, 1, 1, 0,
    ]
], dtype=np.float32)
digits_labels = np.arange(10)


def load_digits():
    """
    Returns dataset that contains discrete digits.

    Returns
    -------
    tuple
        Tuple contains two values. First one is a matrix
        with shape (10, 24). Second one is a vector that contains
        labels for each row. Each digit can be transformed
        into (6, 4) binary image.
    """
    return digits_data, digits_labels


def make_digits(n_samples=100, noise_level=0.1, mode='flip'):
    """
    Returns discrete digits dataset.

    Parameters
    ----------
    n_samples : int
        Number of samples. Defaults to ``100``.

    noise_level : float
        Defines level of a discrete noise added to the images.
        Noise level defines probability for the pixel
        to be removed. Value should be in [0, 1) range.
        Defaults to ``0.1``.

    mode : {``remove``, ``flip``}
        This option allow to specify how additional noise will
        modify each image.

        - ``flip`` - Per every randomly selected pixel function flips
          binary value. ``1 -> 0`` and ``0 -> 1``.

        - ``remove`` - Per every randomly selected pixel function checks
          if value equal to ``1`` if it's true that it gets replaced
          with ``0``.

    Returns
    -------
    tuple
        Tuple contains two values. First one is a matrix
        with shape (n_samples, 24). Second one is a vector
        that contains labels for each row. Each digit can
        be transformed into (6, 4) binary image.

    Examples
    --------
    >>> from neupy import datasets, utils
    >>>
    >>> utils.reproducible()
    >>>
    >>> digits, labels = datasets.make_digits(noise_level=0.15)
    >>> digit, label = digits[0], labels[0]
    >>>
    >>> label
    5
    >>>
    >>> digit.reshape((6, 4))
    array([[0, 0, 1, 1],
           [1, 0, 0, 0],
           [0, 1, 1, 0],
           [0, 0, 0, 0],
           [0, 0, 0, 1],
           [1, 1, 1, 0]], dtype=uint8)
    """
    if mode not in {'remove', 'flip'}:
        raise ValueError("Unknown mode: {}".format(mode))

    if not 0 <= noise_level < 1:
        raise ValueError("noise_level should be float number "
                         "from [0, 1) range, got {}".format(noise_level))

    if n_samples < 1:
        raise ValueError("Number of samples should be an integer greater "
                         "or equal to 1, got {}".format(n_samples))

    digit_indices = np.random.randint(10, size=n_samples)
    digit_images = digits_data[digit_indices]
    digit_labels = digits_labels[digit_indices]

    mask = np.random.binomial(
        n=1, p=noise_level, size=digit_images.shape).astype(bool)

    if mode == 'flip':
        digit_images[mask] = np.where(digit_images[mask] == 1, 0, 1)

    elif mode == 'remove':
        digit_images[mask] = 0

    return digit_images, digit_labels



import tensorflow as tf

from neupy.utils import asfloat
from neupy.exceptions import LayerConnectionError
from neupy.core.properties import (
    ProperFractionProperty,
    NumberProperty,
    TypedListProperty,
)
from .base import Identity


__all__ = ('Dropout', 'GaussianNoise', 'DropBlock')


def bernoulli_sample(mean, shape):
    samples = tf.random_uniform(shape, minval=0, maxval=1, dtype=tf.float32)
    sign_samples = tf.sign(mean - samples)
    return (sign_samples + 1) / 2


class Dropout(Identity):
    """
    Dropout layer. It randomly switches of (multiplies by zero)
    input values, where probability to be switched per each value
    can be controlled with the ``proba`` parameter. For example,
    ``proba=0.2`` will mean that only 20% of the input values will
    be multiplied by 0 and 80% of the will be unchanged.

    It's important to note that output from the dropout is controlled by
    the ``training`` parameter in the ``output`` method. Dropout
    will be applied only in cases when ``training=True`` propagated
    through the network, otherwise it will act as an identity.

    Parameters
    ----------
    proba : float
        Fraction of the input units to drop. Value needs to be
        between ``0`` and ``1``.

    {Identity.name}

    Methods
    -------
    {Identity.Methods}

    Attributes
    ----------
    {Identity.Attributes}

    See Also
    --------
    :layer:`DropBlock` : DropBlock layer.

    References
    ----------
    [1]  Nitish Srivastava, Geoffrey Hinton, Alex Krizhevsky, Ilya Sutskever,
         Ruslan Salakhutdinov, Dropout: a simple way to prevent neural
         networks from overfitting, 2014.

    Examples
    --------
    >>> from neupy.layers import *
    >>> network = join(
    ...     Input(10),
    ...     Relu(5) >> Dropout(0.5),
    ...     Relu(5) >> Dropout(0.5),
    ...     Sigmoid(1),
    ... )
    >>> network
    (?, 10) -> [... 6 layers ...] -> (?, 1)
    """
    proba = ProperFractionProperty()

    def __init__(self, proba, name=None):
        super(Dropout, self).__init__(name=name)
        self.proba = proba

    def output(self, input_value, training=False):
        if not training:
            return input_value
        return tf.nn.dropout(input_value, keep_prob=(1.0 - self.proba))


class GaussianNoise(Identity):
    """
    Add gaussian noise to the input value. Mean and standard deviation
    of the noise can be controlled from the layers parameters.

    It's important to note that output from the layer is controled by
    the ``training`` parameter in the ``output`` method. Layer
    will be applied only in cases when ``training=True`` propagated
    through the network, otherwise it will act as an identity.

    Parameters
    ----------
    std : float
        Standard deviation of the gaussian noise. Values needs to
        be greater than zero. Defaults to ``1``.

    mean : float
        Mean of the gaussian noise. Defaults to ``0``.

    {Identity.name}

    Methods
    -------
    {Identity.Methods}

    Attributes
    ----------
    {Identity.Attributes}

    Examples
    --------
    >>> from neupy.layers import *
    >>> network = join(
    ...     Input(10),
    ...     Relu(5) >> GaussianNoise(std=0.1),
    ...     Relu(5) >> GaussianNoise(std=0.1),
    ...     Sigmoid(1),
    ... )
    >>> network
    (?, 10) -> [... 6 layers ...] -> (?, 1)
    """
    mean = NumberProperty()
    std = NumberProperty(minval=0)

    def __init__(self, mean=1, std=0, name=None):
        super(GaussianNoise, self).__init__(name=name)
        self.mean = mean
        self.std = std

    def output(self, input_value, training=False):
        if not training:
            return input_value

        noise = tf.random_normal(
            shape=tf.shape(input_value),
            mean=self.mean,
            stddev=self.std)

        return input_value + noise


class DropBlock(Identity):
    """
    DropBlock, a form of structured dropout, where units in a contiguous
    region of a feature map are dropped together.

    Parameters
    ----------
    keep_proba : float
        Fraction of the input units to keep. Value needs to be
        between ``0`` and ``1``.

    block_size : int or tuple
        Size of the block to be dropped. Blocks that have squared shape can
        be specified with a single integer value. For example, `block_size=5`
        the same as `block_size=(5, 5)`.

    {Identity.name}

    Methods
    -------
    {Identity.Methods}

    Attributes
    ----------
    {Identity.Attributes}

    See Also
    --------
    :layer:`Dropout` : Dropout layer.

    References
    ----------
    [1] Golnaz Ghiasi, Tsung-Yi Lin, Quoc V. Le. DropBlock: A regularization
        method for convolutional networks, 2018.

    Examples
    --------
    >>> from neupy.layers import *
    >>> network = join(
    ...     Input((28, 28, 1)),
    ...
    ...     Convolution((3, 3, 16)) >> Relu(),
    ...     DropBlock(keep_proba=0.1, block_size=5),
    ...
    ...     Convolution((3, 3, 32)) >> Relu(),
    ...     DropBlock(keep_proba=0.1, block_size=5),
    ... )
    """
    keep_proba = ProperFractionProperty()
    block_size = TypedListProperty(n_elements=2)

    def __init__(self, keep_proba, block_size, name=None):
        super(DropBlock, self).__init__(name=name)

        if isinstance(block_size, int):
            block_size = (block_size, block_size)

        self.keep_proba = keep_proba
        self.block_size = block_size

    def get_output_shape(self, input_shape):
        input_shape = tf.TensorShape(input_shape)

        if input_shape and input_shape.ndims != 4:
            raise LayerConnectionError(
                "DropBlock layer expects input with 4 dimensions, got {} "
                "with shape {}".format(len(input_shape), input_shape))

        return input_shape

    def output(self, input, training=False):
        if not training:
            return input

        input = tf.convert_to_tensor(input, tf.float32)
        input_shape = tf.shape(input)

        block_height, block_width = self.block_size
        height, width = input_shape[1], input_shape[2]

        input_area = asfloat(width * height)
        block_area = asfloat(block_width * block_height)
        area = asfloat((width - block_width + 1) * (height - block_height + 1))

        mask = bernoulli_sample(
            mean=(1. - self.keep_proba) * input_area / (block_area * area),
            shape=[
                input_shape[0],
                height - block_height + 1,
                width - block_width + 1,
                input_shape[3],
            ],
        )

        br_height = (block_height - 1) // 2
        tl_height = (block_height - 1) - br_height

        br_width = (block_width - 1) // 2
        tl_width = (block_width - 1) - br_width

        mask = tf.pad(mask, [
            [0, 0],
            [tl_height, br_height],
            [tl_width, br_width],
            [0, 0],
        ])
        mask = tf.nn.max_pool(
            mask,
            [1, block_height, block_width, 1],
            strides=[1, 1, 1, 1],
            padding='SAME',
        )
        mask = tf.cast(1 - mask, tf.float32)

        feature_normalizer = asfloat(tf.size(mask)) / tf.reduce_sum(mask)
        return tf.multiply(input, mask) * feature_normalizer



from __future__ import division

import tensorflow as tf

from neupy.core.docs import shared_docs
from neupy.utils import asfloat, function_name_scope


__all__ = (
    'mse', 'rmse', 'mae', 'msle', 'rmsle',
    'binary_crossentropy', 'categorical_crossentropy',
    'binary_hinge', 'categorical_hinge',
)


smallest_positive_number = 1e-7  # for 32-bit float numbers


def error_function(expected, predicted):
    """
    Parameters
    ----------
    expected : array-like, tensorflow variable
    predicted : array-like, tensorflow variable

    Returns
    -------
    array-like, tensorflow variable
    """
    raise NotImplementedError


@function_name_scope
@shared_docs(error_function)
def mse(expected, predicted):
    """
    Mean squared error.

    .. math::
        mse(t, o) = mean((t - o) ^ 2)

    where :math:`t=expected` and :math:`o=predicted`

    Parameters
    ----------
    {error_function.expected}
    {error_function.predicted}

    Returns
    -------
    {error_function.Returns}
    """
    return tf.reduce_mean(tf.square(predicted - expected))


@function_name_scope
@shared_docs(error_function)
def rmse(expected, predicted):
    """
    Root mean squared error.

    .. math::
        rmse(t, o) = \\sqrt{{mean((t - o) ^ 2)}} = \\sqrt{{mse(t, 0)}}

    where :math:`t=expected` and :math:`o=predicted`

    Parameters
    ----------
    {error_function.expected}
    {error_function.predicted}

    Returns
    -------
    {error_function.Returns}
    """
    return tf.sqrt(mse(expected, predicted))


@function_name_scope
@shared_docs(error_function)
def mae(expected, predicted):
    """
    Mean absolute error.

    .. math::
        mae(t, o) = mean(\\left| t - o \\right|)

    where :math:`t=expected` and :math:`o=predicted`

    Parameters
    ----------
    {error_function.expected}
    {error_function.predicted}

    Returns
    -------
    {error_function.Returns}
    """
    return tf.reduce_mean(tf.abs(expected - predicted))


@function_name_scope
@shared_docs(error_function)
def msle(expected, predicted):
    """
    Mean squared logarithmic error.

    .. math::
        msle(t, o) = mean((\\log(t + 1) - \\log(o + 1)) ^ 2)

    where :math:`t=expected` and :math:`o=predicted`

    Parameters
    ----------
    {error_function.expected}
    {error_function.predicted}

    Returns
    -------
    {error_function.Returns}
    """
    squared_log = tf.square(tf.log(predicted + 1) - tf.log(expected + 1))
    return tf.reduce_mean(squared_log)


@function_name_scope
@shared_docs(error_function)
def rmsle(expected, predicted):
    """
    Root mean squared logarithmic error.

    .. math::
        rmsle(t, o) = \\sqrt{{
            mean((\\log(t + 1) - \\log(o + 1)) ^ 2)
        }} = \\sqrt{{msle(t, o)}}

    where :math:`t=expected` and :math:`o=predicted`

    Parameters
    ----------
    {error_function.expected}
    {error_function.predicted}

    Returns
    -------
    {error_function.Returns}
    """
    return tf.sqrt(msle(expected, predicted))


@function_name_scope
@shared_docs(error_function)
def binary_crossentropy(expected, predicted):
    """
    Binary cross-entropy error.

    .. math::
        crossentropy(t, o) = -(t\\cdot log(o) + (1 - t) \\cdot log(1 - o))

    where :math:`t=expected` and :math:`o=predicted`

    Parameters
    ----------
    {error_function.expected}
    {error_function.predicted}

    Returns
    -------
    {error_function.Returns}
    """
    epsilon = smallest_positive_number

    predicted = tf.clip_by_value(predicted, epsilon, 1.0 - epsilon)
    crossentropy = (
        -expected * tf.log(predicted) - (1 - expected) * tf.log(1 - predicted))

    return tf.reduce_mean(crossentropy)


@function_name_scope
@shared_docs(error_function)
def categorical_crossentropy(expected, predicted):
    """
    Categorical cross-entropy error.

    Parameters
    ----------
    {error_function.expected}
    {error_function.predicted}

    Returns
    -------
    {error_function.Returns}
    """
    epsilon = smallest_positive_number

    predicted = tf.clip_by_value(predicted, epsilon, 1.0 - epsilon)
    errors = tf.reduce_sum(expected * tf.log(predicted), axis=-1)

    return -tf.reduce_mean(errors)


@function_name_scope
def binary_hinge(expected, predicted, delta=1):
    """
    Computes the binary hinge loss between predictions
    and targets.

    .. math::
        hinge(t, o) = \\max(0, \\delta - t o)

    where :math:`t=expected` and :math:`o=predicted`

    Parameters
    ----------
    expected : Tensorfow tensor
        Targets in {-1, 1} such as ground truth labels.

    predicted : Tensorfow tensor
        Predictions in (-1, 1), such as hyperbolic tangent
        output of a neural network.

    delta : scalar
        The hinge loss margin. Defaults to ``1``.

    Returns
    -------
    Tensorfow tensor
        An expression for the average binary hinge loss.

    Notes
    -----
    This is an alternative to the binary cross-entropy
    loss for binary classification problems.
    """
    error = tf.nn.relu(delta - predicted * expected)
    return tf.reduce_mean(error)


@function_name_scope
def categorical_hinge(expected, predicted, delta=1):
    """
    Computes the multi-class hinge loss between
    predictions and targets.

    .. math::
        hinge_{i}(t, o) = \\max_{j \\not = o_i} (0, t_j - t_{o_i} + \\delta)

    Parameters
    ----------
    expected : Tensorfow 2D tensor or 1D tensor
        Either a vector of int giving the correct class index
        per data point or a 2D tensor of one-hot encoding of
        the correct class in the same layout as predictions
        (non-binary targets in [0, 1] do not work!).

    predicted : Tensorfow 2D tensor
        Predictions in (0, 1), such as softmax output of
        a neural network, with data points in rows and class
        probabilities in columns.

    delta : scalar
        The hinge loss margin. Defaults to ``1``.

    Returns
    -------
    Tensorfow 1D tensor
        An expression for the average multi-class hinge loss.

    Notes
    -----
    This is an alternative to the categorical cross-entropy
    loss for multi-class classification problems.
    """
    shape = tf.shape(expected)
    n_samples = asfloat(shape[0])

    positive = tf.reduce_sum(expected * predicted, axis=-1)
    negative = tf.reduce_max((asfloat(1) - expected) * predicted, axis=-1)
    errors = tf.nn.relu(negative - positive + asfloat(1))
    return tf.reduce_sum(errors) / n_samples



import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from sklearn import model_selection, datasets

from neupy import algorithms
from neupy.layers import *
from neupy.exceptions import LayerConnectionError


def load_data():
    X, y = datasets.fetch_openml('mnist_784', version=1, return_X_y=True)

    X = X / 255.
    X -= X.mean(axis=0)

    x_train, x_test = model_selection.train_test_split(
        X.astype(np.float32),
        test_size=(1 / 7.)
    )
    return x_train, x_test


def generate_and_plot_sampels(network, nx, ny):
    space_x = np.linspace(-3, 3, nx)
    space_y = np.linspace(-3, 3, ny)

    plt.figure(figsize=(12, 12))

    grid = gridspec.GridSpec(nx, ny)
    grid.update(wspace=0, hspace=0)

    for i, x in enumerate(space_x):
        for j, y in enumerate(space_y):
            sample = np.array([[x, y]])
            digit = generator.predict(sample)
            image = digit[0].reshape((28, 28))

            plt.subplot(grid[nx * i + j])
            plt.imshow(image, cmap='Greys_r')
            plt.axis('off')

    plt.show()


class GaussianSample(BaseLayer):
    def get_output_shape(self, mu_shape, sigma_shape):
        return mu_shape

    def output(self, mu, sigma, **kwargs):
        return mu + tf.exp(sigma) * tf.random_normal(tf.shape(mu))


class Collect(Identity):
    def output(self, input, **kwargs):
        tf.add_to_collection(self.name, input)
        return input


def binary_crossentropy(expected, predicted):
    epsilon = 1e-7  # smallest positive 32-bit float number
    predicted = tf.clip_by_value(predicted, epsilon, 1.0 - epsilon)

    return tf.reduce_sum(
        expected * tf.log(predicted) +
        (1 - expected) * tf.log(1 - predicted),
        axis=1
    )


def vae_loss(expected, predicted):
    mean = tf.get_collection('mu')[-1]
    log_var = tf.get_collection('sigma')[-1]

    epsilon = 1e-7
    predicted = tf.clip_by_value(predicted, epsilon, 1.0 - epsilon)

    crossentropy_loss = binary_crossentropy(expected, predicted)
    kl_loss = tf.reduce_sum(
        1 + log_var - tf.square(mean) - tf.exp(log_var),
        axis=1
    )
    return tf.reduce_mean(-crossentropy_loss - 0.5 * kl_loss)


if __name__ == '__main__':
    # Construct Variational Autoencoder
    network = join(
        # Encoder
        Input(784, name='input'),
        Tanh(256),

        # Sampler
        parallel(
            # Two is the maximum number of dimensions that we can visualize
            Linear(2, name='mu') >> Collect('mu'),
            Linear(2, name='sigma') >> Collect('sigma'),
        ),
        GaussianSample(),

        # Decoder
        # Note: Identity layer acts as a reference. Using it
        # we can easily cut the decoder from the network
        Identity('decoder'),
        Tanh(256),
        Sigmoid(784),
    )

    # Train network
    optimizer = algorithms.RMSProp(
        network,
        loss=vae_loss,
        regularizer=algorithms.l2(0.001),

        batch_size=128,
        shuffle_data=True,
        step=0.001,
        verbose=True,
    )

    x_train, x_test = load_data()
    optimizer.train(x_train, x_train, x_test, x_test, epochs=50)

    # Sample digits from the obtained distribution
    generator = Input(2) >> network.start('decoder')
    generate_and_plot_sampels(generator, 15, 15)



import numpy as np
import tensorflow as tf
from scipy.sparse import issparse


__all__ = ('format_data', 'asfloat')


def format_data(data, is_feature1d=True, copy=False, make_float=True):
    """
    Transform data in a standardized format.

    Notes
    -----
    It should be applied to the input data prior to use in
    learning algorithms.

    Parameters
    ----------
    data : array-like
        Data that should be formatted. That could be, matrix, vector or
        Pandas DataFrame instance.

    is_feature1d : bool
        Should be equal to ``True`` if input data is a vector that
        contains N samples with 1 feature each. Defaults to ``True``.

    copy : bool
        Defaults to ``False``.

    make_float : bool
        If `True` then input will be converted to float.
        Defaults to ``False``.

    Returns
    -------
    ndarray
        The same input data but transformed to a standardized format
        for further use.
    """
    if data is None or issparse(data):
        return data

    if make_float:
        data = asfloat(data)

    if not isinstance(data, np.ndarray) or copy:
        data = np.array(data, copy=copy)

    # Valid number of features for one or two dimensions
    n_features = data.shape[-1]

    if data.ndim == 1:
        data_shape = (n_features, 1) if is_feature1d else (1, n_features)
        data = data.reshape(data_shape)

    return data


def asfloat(value):
    """
    Convert variable to 32 bit float number.

    Parameters
    ----------
    value : matrix, ndarray, Tensorfow variable or scalar
        Value that could be converted to float type.

    Returns
    -------
    matrix, ndarray, Tensorfow variable or scalar
        Output would be input value converted to 32 bit float.
    """
    float_type = 'float32'

    if isinstance(value, (np.matrix, np.ndarray)):
        if value.dtype != np.dtype(float_type):
            return value.astype(float_type)

        return value

    elif isinstance(value, (tf.Tensor, tf.SparseTensor)):
        return tf.cast(value, tf.float32)

    elif issparse(value):
        return value

    float_x_type = np.cast[float_type]
    return float_x_type(value)



import numpy as np
import tensorflow as tf

from neupy.core.config import DumpableObject
from neupy.core.properties import IntProperty, ParameterProperty
from neupy.algorithms.base import BaseNetwork
from neupy.utils import asfloat, format_data, dot, iters, tf_utils
from neupy import init


__all__ = ('RBM',)


def random_binomial(p):
    with tf.name_scope('random-binomial'):
        samples = tf.random_uniform(tf.shape(p), dtype=tf.float32) <= p
        return tf.cast(samples, tf.float32)


def random_sample(data, n_samples):
    with tf.name_scope('random-sample'):
        data_shape = tf.shape(data)
        max_index = data_shape[0]
        sample_indices = tf.random_uniform(
            (n_samples,), minval=0, maxval=max_index,
            dtype=tf.int32
        )
        return tf.gather(data, sample_indices)


class RBM(BaseNetwork, DumpableObject):
    """
    Boolean/Bernoulli Restricted Boltzmann Machine (RBM).
    Algorithm assumes that inputs are either binary
    values or values between 0 and 1.

    Parameters
    ----------
    n_visible : int
        Number of visible units. Number of features (columns)
        in the input data.

    n_hidden : int
        Number of hidden units. The large the number the more
        information network can capture from the data, but it
        also mean that network is more likely to overfit.

    batch_size : int
        Size of the mini-batch. Defaults to ``10``.

    weight : array-like, Tensorfow variable, Initializer or scalar
        Default initialization methods
        you can find :ref:`here <init-methods>`.
        Defaults to :class:`Normal <neupy.init.Normal>`.

    hidden_bias : array-like, Tensorfow variable, Initializer or scalar
        Default initialization methods
        you can find :ref:`here <init-methods>`.
        Defaults to :class:`Constant(value=0) <neupy.init.Constant>`.

    visible_bias : array-like, Tensorfow variable, Initializer or scalar
        Default initialization methods
        you can find :ref:`here <init-methods>`.
        Defaults to :class:`Constant(value=0) <neupy.init.Constant>`.

    {BaseNetwork.Parameters}

    Methods
    -------
    train(X_train, epochs=100)
        Trains network.

    predict(X)
        Alias to the ``visible_to_hidden`` method.

    visible_to_hidden(visible_input)
        Propagates data through the network and returns output
        from the hidden layer.

    hidden_to_visible(hidden_input)
        Propagates output from the hidden layer backward
        to the visible.

    gibbs_sampling(visible_input, n_iter=1)
        Makes Gibbs sampling ``n`` times using visible input.

    {BaseSkeleton.fit}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> data = np.array([
    ...     [1, 0, 1, 0],
    ...     [1, 0, 1, 0],
    ...     [1, 0, 0, 0],  # incomplete sample
    ...     [1, 0, 1, 0],
    ...
    ...     [0, 1, 0, 1],
    ...     [0, 0, 0, 1],  # incomplete sample
    ...     [0, 1, 0, 1],
    ...     [0, 1, 0, 1],
    ...     [0, 1, 0, 1],
    ...     [0, 1, 0, 1],
    ... ])
    >>>
    >>> rbm = algorithms.RBM(n_visible=4, n_hidden=1)
    >>> rbm.train(data, epochs=100)
    >>>
    >>> hidden_states = rbm.visible_to_hidden(data)
    >>> hidden_states.round(2)
    array([[ 0.99],
           [ 0.99],
           [ 0.95],
           [ 0.99],
           [ 0.  ],
           [ 0.01],
           [ 0.  ],
           [ 0.  ],
           [ 0.  ],
           [ 0.  ]])

    References
    ----------
    [1] G. Hinton, A Practical Guide to Training Restricted
        Boltzmann Machines, 2010.
        http://www.cs.toronto.edu/~hinton/absps/guideTR.pdf
    """
    n_visible = IntProperty(minval=1)
    n_hidden = IntProperty(minval=1)
    batch_size = IntProperty(minval=1, default=10)

    weight = ParameterProperty(default=init.Normal())
    hidden_bias = ParameterProperty(default=init.Constant(value=0))
    visible_bias = ParameterProperty(default=init.Constant(value=0))

    def __init__(self, n_visible, n_hidden, **options):
        options.update({'n_visible': n_visible, 'n_hidden': n_hidden})
        super(RBM, self).__init__(**options)

        self.init_functions()
        self.init_methods()

    def init_functions(self):
        with tf.variable_scope('rbm'):
            self.weight = tf_utils.create_variable(
                value=self.weight,
                name='weight',
                shape=(self.n_visible, self.n_hidden),
                trainable=False,
            )
            self.hidden_bias = tf_utils.create_variable(
                value=self.hidden_bias,
                name='hidden-bias',
                shape=(self.n_hidden,),
                trainable=False,
            )
            self.visible_bias = tf_utils.create_variable(
                value=self.visible_bias,
                name='visible-bias',
                shape=(self.n_visible,),
                trainable=False,
            )
            self.h_samples = tf.Variable(
                tf.zeros([self.batch_size, self.n_hidden]),
                name="hidden-samples",
                dtype=tf.float32,
                trainable=False,
            )

            self.network_input = tf.placeholder(
                tf.float32,
                (None, self.n_visible),
                name="network-input",
            )
            self.network_hidden_input = tf.placeholder(
                tf.float32,
                (None, self.n_hidden),
                name="network-hidden-input",
            )

    def init_methods(self):
        def free_energy(visible_sample):
            with tf.name_scope('free-energy'):
                wx = tf.matmul(visible_sample, self.weight)
                wx_b = wx + self.hidden_bias

                visible_bias_term = dot(visible_sample, self.visible_bias)

                # We can get infinity when wx_b is a relatively large number
                # (maybe 100). Taking exponent makes it even larger and
                # for with float32 it can convert it to infinity. But because
                # number is so large we don't care about +1 value before taking
                # logarithms and therefore we can just pick value as it is
                # since our operation won't change anything.
                hidden_terms = tf.where(
                    # exp(30) is such a big number that +1 won't
                    # make any difference in the outcome.
                    tf.greater(wx_b, 30),
                    wx_b,
                    tf.log1p(tf.exp(wx_b)),
                )

                hidden_term = tf.reduce_sum(hidden_terms, axis=1)
                return -(visible_bias_term + hidden_term)

        def visible_to_hidden(visible_sample):
            with tf.name_scope('visible-to-hidden'):
                wx = tf.matmul(visible_sample, self.weight)
                wx_b = wx + self.hidden_bias
                return tf.nn.sigmoid(wx_b)

        def hidden_to_visible(hidden_sample):
            with tf.name_scope('hidden-to-visible'):
                wx = tf.matmul(hidden_sample, self.weight, transpose_b=True)
                wx_b = wx + self.visible_bias
                return tf.nn.sigmoid(wx_b)

        def sample_hidden_from_visible(visible_sample):
            with tf.name_scope('sample-hidden-to-visible'):
                hidden_prob = visible_to_hidden(visible_sample)
                hidden_sample = random_binomial(hidden_prob)
                return hidden_sample

        def sample_visible_from_hidden(hidden_sample):
            with tf.name_scope('sample-visible-to-hidden'):
                visible_prob = hidden_to_visible(hidden_sample)
                visible_sample = random_binomial(visible_prob)
                return visible_sample

        network_input = self.network_input
        network_hidden_input = self.network_hidden_input
        input_shape = tf.shape(network_input)
        n_samples = input_shape[0]

        weight = self.weight
        h_bias = self.hidden_bias
        v_bias = self.visible_bias
        h_samples = self.h_samples
        step = asfloat(self.step)

        with tf.name_scope('positive-values'):
            # We have to use `cond` instead of `where`, because
            # different if-else cases might have different shapes
            # and it triggers exception in tensorflow.
            v_pos = tf.cond(
                tf.equal(n_samples, self.batch_size),
                lambda: network_input,
                lambda: random_sample(network_input, self.batch_size)
            )
            h_pos = visible_to_hidden(v_pos)

        with tf.name_scope('negative-values'):
            v_neg = sample_visible_from_hidden(h_samples)
            h_neg = visible_to_hidden(v_neg)

        with tf.name_scope('weight-update'):
            weight_update = (
                tf.matmul(v_pos, h_pos, transpose_a=True) -
                tf.matmul(v_neg, h_neg, transpose_a=True)
            ) / asfloat(n_samples)

        with tf.name_scope('hidden-bias-update'):
            h_bias_update = tf.reduce_mean(h_pos - h_neg, axis=0)

        with tf.name_scope('visible-bias-update'):
            v_bias_update = tf.reduce_mean(v_pos - v_neg, axis=0)

        with tf.name_scope('flipped-input-features'):
            # Each row will have random feature marked with number 1
            # Other values will be equal to 0
            possible_feature_corruptions = tf.eye(self.n_visible)
            corrupted_features = random_sample(
                possible_feature_corruptions, n_samples)

            rounded_input = tf.round(network_input)
            # If we scale input values from [0, 1] range to [-1, 1]
            # than it will be easier to flip feature values with simple
            # multiplication.
            scaled_rounded_input = 2 * rounded_input - 1
            scaled_flipped_rounded_input = (
                # for corrupted_features we convert 0 to 1 and 1 to -1
                # in this way after multiplication we will flip all
                # signs where -1 in the transformed corrupted_features
                (-2 * corrupted_features + 1) * scaled_rounded_input
            )
            # Scale it back to the [0, 1] range
            flipped_rounded_input = (scaled_flipped_rounded_input + 1) / 2

        with tf.name_scope('pseudo-likelihood-loss'):
            # Stochastic pseudo-likelihood
            error = tf.reduce_mean(
                self.n_visible * tf.log_sigmoid(
                    free_energy(flipped_rounded_input) -
                    free_energy(rounded_input)
                )
            )

        with tf.name_scope('gibbs-sampling'):
            gibbs_sampling = sample_visible_from_hidden(
                sample_hidden_from_visible(network_input))

        tf_utils.initialize_uninitialized_variables()
        self.weight_update_one_step = tf_utils.function(
            [network_input],
            error,
            name='rbm/train-epoch',
            updates=[
                (weight, weight + step * weight_update),
                (h_bias, h_bias + step * h_bias_update),
                (v_bias, v_bias + step * v_bias_update),
                (h_samples, random_binomial(p=h_neg)),
            ]
        )
        self.score_func = tf_utils.function(
            [network_input],
            error,
            name='rbm/prediction-error',
        )
        self.visible_to_hidden_one_step = tf_utils.function(
            [network_input],
            visible_to_hidden(network_input),
            name='rbm/visible-to-hidden',
        )
        self.hidden_to_visible_one_step = tf_utils.function(
            [network_hidden_input],
            hidden_to_visible(network_hidden_input),
            name='rbm/hidden-to-visible',
        )
        self.gibbs_sampling_one_step = tf_utils.function(
            [network_input],
            gibbs_sampling,
            name='rbm/gibbs-sampling',
        )

    def train(self, X_train, X_test=None, epochs=100):
        """
        Train RBM.

        Parameters
        ----------
        X_train : 1D or 2D array-like
        X_test : 1D or 2D array-like or None
            Defaults to ``None``.
        epochs : int
            Number of training epochs. Defaults to ``100``.
        """
        return super(RBM, self).train(
            X_train=X_train, y_train=None,
            X_test=X_test, y_test=None,
            epochs=epochs)

    def one_training_update(self, X_train, y_train=None):
        """
        Train one epoch.

        Parameters
        ----------
        X_train : array-like (n_samples, n_features)

        Returns
        -------
        float
        """
        return self.weight_update_one_step(X_train)

    def predict(self, X):
        return self.visible_to_hidden(X)

    def visible_to_hidden(self, visible_input):
        """
        Propagates data through the network and returns output
        from the hidden layer.

        Parameters
        ----------
        visible_input : array-like (n_samples, n_visible_features)

        Returns
        -------
        array-like
        """
        is_input_feature1d = (self.n_visible == 1)
        visible_input = format_data(visible_input, is_input_feature1d)

        outputs = iters.apply_batches(
            function=self.visible_to_hidden_one_step,
            inputs=visible_input,
            batch_size=self.batch_size,
            show_progressbar=self.logs.enable,
        )
        return np.concatenate(outputs, axis=0)

    def hidden_to_visible(self, hidden_input):
        """
        Propagates output from the hidden layer backward
        to the visible.

        Parameters
        ----------
        hidden_input : array-like (n_samples, n_hidden_features)

        Returns
        -------
        array-like
        """
        is_input_feature1d = (self.n_hidden == 1)
        hidden_input = format_data(hidden_input, is_input_feature1d)

        outputs = iters.apply_batches(
            function=self.hidden_to_visible_one_step,
            inputs=hidden_input,
            batch_size=self.batch_size,
            show_progressbar=self.logs.enable,
        )
        return np.concatenate(outputs, axis=0)

    def score(self, X, y=None):
        """
        Compute the pseudo-likelihood of input samples.

        Parameters
        ----------
        X : array-like
            Values of the visible layer

        Returns
        -------
        float
            Value of the pseudo-likelihood.
        """
        return iters.apply_batches(
            function=self.score_func,
            inputs=format_data(X, is_feature1d=(self.n_visible == 1)),
            batch_size=self.batch_size,

            show_output=True,
            show_progressbar=self.logs.enable,
            average_outputs=True,
        )

    def gibbs_sampling(self, visible_input, n_iter=1):
        """
        Makes Gibbs sampling n times using visible input.

        Parameters
        ----------
        visible_input : 1d or 2d array
        n_iter : int
            Number of Gibbs sampling iterations. Defaults to ``1``.

        Returns
        -------
        array-like
            Output from the visible units after perfoming n
            Gibbs samples. Array will contain only binary
            units (0 and 1).
        """
        is_input_feature1d = (self.n_visible == 1)
        visible_input = format_data(visible_input, is_input_feature1d)

        input_ = visible_input
        for iteration in range(n_iter):
            input_ = self.gibbs_sampling_one_step(input_)

        return input_



from operator import attrgetter

import numpy as np

from neupy.utils import format_data
from neupy.exceptions import StopTraining
from neupy.algorithms.base import BaseNetwork
from neupy.core.properties import (NumberProperty, ProperFractionProperty,
                                   IntProperty)


__all__ = ('GrowingNeuralGas', 'NeuralGasGraph', 'NeuronNode')


def sample_data_point(data, n=1):
    indices = np.random.choice(len(data), n)
    return data[indices, :]


class NeuralGasGraph(object):
    """
    Undirected graph structure that stores neural gas network's
    neurons and connections between them.

    Attributes
    ----------
    edges_per_node : dict
        Dictionary that where key is a unique node and value is a list
        of nodes that connection to this edge.

    edges : dict
        Dictonary that stores age per each connection. Ech key will have
        the following format: ``(node_1, node_2)``.

    nodes : list
        List of all nodes in the graph (read-only attribute).

    n_nodes : int
        Number of nodes in the network (read-only attribute).

    n_edges : int
        Number of edges in the network (read-only attribute).
    """
    def __init__(self):
        self.edges_per_node = {}
        self.edges = {}

    @property
    def nodes(self):
        return list(self.edges_per_node.keys())

    @property
    def n_nodes(self):
        return len(self.edges_per_node)

    @property
    def n_edges(self):
        return len(self.edges)

    def add_node(self, node):
        self.edges_per_node[node] = set()

    def remove_node(self, node):
        if self.edges_per_node[node]:
            raise ValueError(
                "Cannot remove node, because it's connected to "
                "{} other node(s)".format(len(self.edges_per_node[node])))

        del self.edges_per_node[node]

    def add_edge(self, node_1, node_2):
        if node_2 in self.edges_per_node[node_1]:
            return self.reset_edge(node_1, node_2)

        self.edges_per_node[node_1].add(node_2)
        self.edges_per_node[node_2].add(node_1)
        self.edges[(node_1, node_2)] = 0

    def reset_edge(self, node_1, node_2):
        edge_id = self.find_edge_id(node_1, node_2)
        self.edges[edge_id] = 0

    def remove_edge(self, node_1, node_2):
        edge_id = self.find_edge_id(node_1, node_2)

        self.edges_per_node[node_1].remove(node_2)
        self.edges_per_node[node_2].remove(node_1)

        del self.edges[edge_id]

    def find_edge_id(self, node_1, node_2):
        if (node_1, node_2) in self.edges:
            return (node_1, node_2)

        if (node_2, node_1) in self.edges:
            return (node_2, node_1)

        raise ValueError("Edge between specified nodes doesn't exist")

    def __repr__(self):
        return "<{} n_nodes={}, n_edges={}>".format(
            self.__class__.__name__,
            self.n_nodes, self.n_edges)


class NeuronNode(object):
    """
    Structure representes neuron in the Neural Gas algorithm.

    Attributes
    ----------
    weight : 2d-array
        Neuron's position in the space.

    error : float
        Error accumulated during the training.
    """
    def __init__(self, weight):
        self.weight = weight
        self.error = 0

    def __repr__(self):
        return "<{} error={}>".format(
            self.__class__.__name__,
            round(float(self.error), 6))


class GrowingNeuralGas(BaseNetwork):
    """
    Growing Neural Gas (GNG) algorithm.

    Current algorithm has two modifications that hasn't been mentioned
    in the paper, but they help to speed up training.

    - The ``n_start_nodes`` parameter provides possibility to increase
      number of nodes during initialization step. It's useful when
      algorithm takes a lot of time building up large amount of neurons.

    - The ``min_distance_for_update`` parameter allows to speed up
      training when some data samples has neurons very close to them. The
      ``min_distance_for_update`` parameter controls threshold for the
      minimum distance for which we will want to update weights.

    Parameters
    ----------
    n_inputs : int
        Number of features in each sample.

    n_start_nodes : int
        Number of nodes that algorithm generates from the data during
        the initialization step. Defaults to ``2``.

    step : float
        Step (learning rate) for the neuron winner. Defaults to ``0.2``.

    neighbour_step : float
        Step (learning rate) for the neurons that connected via edges
        with neuron winner. This value typically has to be smaller than
        ``step`` value. Defaults to ``0.05``.

    max_edge_age : int
        It means that if edge won't be updated for ``max_edge_age`` iterations
        than it would be removed. The larger the value the more updates we
        allow to do before removing edge. Defaults to ``100``.

    n_iter_before_neuron_added : int
        Each ``n_iter_before_neuron_added`` weight update algorithm add new
        neuron. The smaller the value the more frequently algorithm adds
        new neurons to the network. Defaults to ``1000``.

    error_decay_rate : float
        This error decay rate would be applied to every neuron in the
        graph after each training iteration. It ensures that old errors
        will be reduced over time. Defaults to ``0.995``.

    after_split_error_decay_rate : float
        This decay rate reduces error for neurons with largest errors
        after algorithm added new neuron. This value typically lower than
        ``error_decay_rate``. Defaults to ``0.5``.

    max_nodes : int
        Maximum number of nodes that would be generated during the training.
        This parameter won't stop training when maximum number of nodes
        will be exceeded. Defaults to ``1000``.

    min_distance_for_update : float
        Parameter controls for which neurons we want to apply updates.
        In case if euclidean distance between data sample and closest
        neurons will be less than the ``min_distance_for_update`` value than
        update would be skipped for this data sample. Setting value to zero
        will disable effect provided by this parameter. Defaults to ``0``.

    {BaseNetwork.show_epoch}

    {BaseNetwork.shuffle_data}

    {BaseNetwork.signals}

    {Verbose.verbose}

    Methods
    -------
    train(X_train, epochs=100)
        Network learns topological structure of the data. Learned
        structure will be stored in the ``graph`` attribute.

    {BaseSkeleton.fit}

    initialize_nodes(data)
        Network initializes nodes randomly sampling ``n_start_nodes``
        from the data. It would be applied automatically before
        the training in case if graph is empty.

        Note: Node re-initialization can reset network.

    Notes
    -----
    - Unlike other algorithms this network doesn't make predictions.
      Instead, it learns topological structure of the data in form of
      the graph. After that training, structure of the network can be
      extracted from the ``graph`` attribute.

    - In order to speed up training, it might be useful to increase
      the ``n_start_nodes`` parameter.

    - During the training it happens that nodes learn topological
      structure of one part of the data better than the other, mostly
      because of the different data sample density in different places.
      Increasing the ``min_distance_for_update`` can speed up training
      ignoring updates for the neurons that very close to the data sample.
      (below specified ``min_distance_for_update`` value). Training can be
      stopped in case if none of the neurons has been updated during
      the training epoch.

    Attributes
    ----------
    graph : NeuralGasGraph instance
        This attribute stores all neurons and connections between them
        in the form of undirected graph.

    {BaseNetwork.Attributes}

    Examples
    --------
    >>> from neupy import algorithms
    >>> from sklearn.datasets import make_blobs
    >>>
    >>> data, _ = make_blobs(
    ...     n_samples=1000,
    ...     n_features=2,
    ...     centers=2,
    ...     cluster_std=0.4,
    ... )
    >>>
    >>> neural_gas = algorithms.GrowingNeuralGas(
    ...     n_inputs=2,
    ...     shuffle_data=True,
    ...     verbose=True,
    ...     max_edge_age=10,
    ...     n_iter_before_neuron_added=50,
    ...     max_nodes=100,
    ... )
    >>> neural_gas.graph.n_nodes
    100
    >>> len(neural_gas.graph.edges)
    175
    >>> edges = list(neural_gas.graph.edges.keys())
    >>> neuron_1, neuron_2 = edges[0]
    >>>
    >>> neuron_1.weight
    array([[-6.77166299,  2.4121606 ]])
    >>> neuron_2.weight
    array([[-6.829309  ,  2.27839633]])

    References
    ----------
    [1] A Growing Neural Gas Network Learns Topologies, Bernd Fritzke
    """
    n_inputs = IntProperty(minval=1, required=True)
    n_start_nodes = IntProperty(minval=2, default=2)

    step = NumberProperty(default=0.2, minval=0)
    neighbour_step = NumberProperty(default=0.05, minval=0)
    max_edge_age = IntProperty(default=100, minval=1)
    max_nodes = IntProperty(default=1000, minval=1)

    n_iter_before_neuron_added = IntProperty(default=1000, minval=1)
    after_split_error_decay_rate = ProperFractionProperty(default=0.5)
    error_decay_rate = ProperFractionProperty(default=0.995)
    min_distance_for_update = NumberProperty(default=0.0, minval=0)

    def __init__(self, *args, **kwargs):
        super(GrowingNeuralGas, self).__init__(*args, **kwargs)
        self.n_updates = 0
        self.graph = NeuralGasGraph()

    def format_input_data(self, X):
        is_feature1d = self.n_inputs == 1
        X = format_data(X, is_feature1d)

        if X.ndim != 2:
            raise ValueError("Cannot make prediction, because input "
                             "data has more than 2 dimensions")

        n_samples, n_features = X.shape

        if n_features != self.n_inputs:
            raise ValueError("Input data expected to have {} features, "
                             "but got {}".format(self.n_inputs, n_features))

        return X

    def initialize_nodes(self, data):
        self.graph = NeuralGasGraph()

        for sample in sample_data_point(data, n=self.n_start_nodes):
            self.graph.add_node(NeuronNode(sample.reshape(1, -1)))

    def train(self, X_train, epochs=100):
        X_train = self.format_input_data(X_train)

        if not self.graph.nodes:
            self.initialize_nodes(X_train)

        return super(GrowingNeuralGas, self).train(
            X_train=X_train, y_train=None,
            X_test=None, y_test=None,
            epochs=epochs)

    def one_training_update(self, X_train, y_train=None):
        graph = self.graph
        step = self.step
        neighbour_step = self.neighbour_step

        max_nodes = self.max_nodes
        max_edge_age = self.max_edge_age

        error_decay_rate = self.error_decay_rate
        after_split_error_decay_rate = self.after_split_error_decay_rate
        n_iter_before_neuron_added = self.n_iter_before_neuron_added

        # We square this value, because we deal with
        # squared distances during the training.
        min_distance_for_update = np.square(self.min_distance_for_update)

        n_samples = len(X_train)
        total_error = 0
        did_update = False

        for sample in X_train:
            nodes = graph.nodes
            weights = np.concatenate([node.weight for node in nodes])

            distance = np.linalg.norm(weights - sample, axis=1)
            neuron_ids = np.argsort(distance)

            closest_neuron_id, second_closest_id = neuron_ids[:2]
            closest_neuron = nodes[closest_neuron_id]
            second_closest = nodes[second_closest_id]
            total_error += distance[closest_neuron_id]

            if distance[closest_neuron_id] < min_distance_for_update:
                continue

            self.n_updates += 1
            did_update = True

            closest_neuron.error += distance[closest_neuron_id]
            closest_neuron.weight += step * (sample - closest_neuron.weight)

            graph.add_edge(closest_neuron, second_closest)

            for to_neuron in list(graph.edges_per_node[closest_neuron]):
                edge_id = graph.find_edge_id(to_neuron, closest_neuron)
                age = graph.edges[edge_id]

                if age >= max_edge_age:
                    graph.remove_edge(to_neuron, closest_neuron)

                    if not graph.edges_per_node[to_neuron]:
                        graph.remove_node(to_neuron)

                else:
                    graph.edges[edge_id] += 1
                    to_neuron.weight += neighbour_step * (
                        sample - to_neuron.weight)

            time_to_add_new_neuron = (
                self.n_updates % n_iter_before_neuron_added == 0 and
                graph.n_nodes < max_nodes)

            if time_to_add_new_neuron:
                nodes = graph.nodes
                largest_error_neuron = max(nodes, key=attrgetter('error'))
                neighbour_neuron = max(
                    graph.edges_per_node[largest_error_neuron],
                    key=attrgetter('error'))

                largest_error_neuron.error *= after_split_error_decay_rate
                neighbour_neuron.error *= after_split_error_decay_rate

                new_weight = 0.5 * (
                    largest_error_neuron.weight + neighbour_neuron.weight
                )
                new_neuron = NeuronNode(weight=new_weight.reshape(1, -1))

                graph.remove_edge(neighbour_neuron, largest_error_neuron)
                graph.add_node(new_neuron)
                graph.add_edge(largest_error_neuron, new_neuron)
                graph.add_edge(neighbour_neuron, new_neuron)

            for node in graph.nodes:
                node.error *= error_decay_rate

        if not did_update and min_distance_for_update != 0 and n_samples > 1:
            raise StopTraining(
                "Distance between every data sample and neurons, closest "
                "to them, is less then {}".format(min_distance_for_update))

        return total_error / n_samples

    def predict(self, *args, **kwargs):
        raise NotImplementedError(
            "Growing Neural Gas algorithm doesn't make prediction. "
            "It only learns graph structure from the data "
            "(class has `graph` attribute). ")
